"""Subclass of :class:`click.Group` that loads subcommands dynamically from entry points."""

from __future__ import annotations

import functools
import re
import typing as t

import click

from aiida.common import exceptions
from aiida.plugins.entry_point import ENTRY_POINT_GROUP_FACTORY_MAPPING, get_entry_point_names
from aiida.plugins.factories import BaseFactory

from ..params import options
from ..params.options.interactive import InteractiveOption
from .verdi import VerdiCommandGroup

__all__ = ('DynamicEntryPointCommandGroup',)


class DynamicEntryPointCommandGroup(VerdiCommandGroup):
    """Subclass of :class:`click.Group` that loads subcommands dynamically from entry points.

    A command group using this class will automatically generate the sub commands from the entry points registered in
    the given ``entry_point_group``. The entry points can be additionally filtered using a regex defined for the
    ``entry_point_name_filter`` keyword. The actual command for each entry point is defined by ``command``, which should
    take as a first argument the class that corresponds to the entry point. In addition, it should accept ``kwargs``
    which will be the values for the options passed when the command is invoked. The help string of the command will be
    provided by the docstring of the class registered at the respective entry point. Example usage:

    .. code:: python

        def create_instance(cls, **kwargs):
            instance = cls(**kwargs)
            instance.store()
            echo.echo_success(f'Created {cls.__name__}<{instance.pk}>')

        @click.group('create', cls=DynamicEntryPointCommandGroup, command=create_instance,)
        def cmd_create():
            pass

    """

    def __init__(
        self,
        command: t.Callable,
        entry_point_group: str,
        entry_point_name_filter: str = r'.*',
        shared_options: list[click.Option] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._command = command
        self.entry_point_group = entry_point_group
        self.entry_point_name_filter = entry_point_name_filter
        self.factory = ENTRY_POINT_GROUP_FACTORY_MAPPING.get(
            entry_point_group, functools.partial(BaseFactory, entry_point_group)
        )
        self.shared_options = shared_options

    def list_commands(self, ctx: click.Context) -> list[str]:
        """Return the sorted list of subcommands for this group.

        :param ctx: The :class:`click.Context`.
        """
        commands = super().list_commands(ctx)
        commands.extend(
            [
                entry_point
                for entry_point in get_entry_point_names(self.entry_point_group)
                if re.match(self.entry_point_name_filter, entry_point)
                and getattr(self.factory(entry_point), 'cli_exposed', True)
            ]
        )
        return sorted(commands)

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        """Return the command with the given name.

        :param ctx: The :class:`click.Context`.
        :param cmd_name: The name of the command.
        :returns: The :class:`click.Command`.
        """
        try:
            command: click.Command | None = self.create_command(ctx, cmd_name)
        except exceptions.EntryPointError:
            command = super().get_command(ctx, cmd_name)
        return command

    def call_command(self, ctx, cls, **kwargs):
        """Call the ``command`` after validating the provided inputs."""
        from pydantic import ValidationError

        if hasattr(cls, 'Model'):
            # The plugin defines a pydantic model: use it to validate the provided arguments
            try:
                model = cls.Model(**kwargs)
            except ValidationError as exception:
                param_hint = [
                    f'--{loc.replace("_", "-")}'  # type: ignore[union-attr]
                    for loc in exception.errors()[0]['loc']
                ]
                message = '\n'.join([str(e['ctx']['error']) for e in exception.errors()])
                raise click.BadParameter(
                    message,
                    param_hint=param_hint or 'multiple parameters',  # type: ignore[arg-type]
                ) from exception

            # Update the arguments with the dictionary representation of the model. This will include any type coercions
            # that may have been applied with validators defined for the model.
            kwargs.update(**model.model_dump())

        return self._command(ctx, cls, **kwargs)

    def create_command(self, ctx: click.Context, entry_point: str) -> click.Command:
        """Create a subcommand for the given ``entry_point``."""
        cls = self.factory(entry_point)
        command = functools.partial(self.call_command, ctx, cls)
        command.__doc__ = cls.__doc__
        return click.command(entry_point)(self.create_options(entry_point)(command))

    def create_options(self, entry_point: str) -> t.Callable:
        """Create the option decorators for the command function for the given entry point.

        :param entry_point: The entry point.
        """

        def apply_options(func):
            """Decorate the command function with the appropriate options for the given entry point."""
            func = options.NON_INTERACTIVE()(func)
            func = options.CONFIG_FILE()(func)

            options_list = self.list_options(entry_point)
            options_list.reverse()

            for option in options_list:
                func = option(func)

            shared_options = self.shared_options or []
            shared_options.reverse()

            for option in shared_options:
                func = option(func)

            return func

        return apply_options

    def list_options(self, entry_point: str) -> list:
        """Return the list of options that should be applied to the command for the given entry point.

        :param entry_point: The entry point.
        """
        from pydantic_core import PydanticUndefined

        cls = self.factory(entry_point)

        if not hasattr(cls, 'Model'):
            from aiida.common.warnings import warn_deprecation

            warn_deprecation(
                'Relying on `_get_cli_options` is deprecated. The options should be defined through a '
                '`pydantic.BaseModel` that should be assigned to the `Model` class attribute.',
                version=3,
            )
            options_spec = self.factory(entry_point).get_cli_options()  # type: ignore[union-attr]
            return [self.create_option(*item) for item in options_spec]

        options_spec = {}

        for key, field_info in cls.Model.model_fields.items():
            default = field_info.default_factory if field_info.default is PydanticUndefined else field_info.default

            # If the annotation has the ``__args__`` attribute it is an instance of a type from ``typing`` and the real
            # type can be gotten from the arguments. For example it could be ``typing.Union[str, None]`` calling
            # ``typing.Union[str, None].__args__`` will return the tuple ``(str, NoneType)``. So to get the real type,
            # we simply remove all ``NoneType`` and the remaining type should be the type of the option.
            if hasattr(field_info.annotation, '__args__'):
                args = list(filter(lambda e: e != type(None), field_info.annotation.__args__))
                # Click parameters only support specifying a single type, so we default to the first one even if the
                # pydantic model defines multiple.
                field_type = args[0]
            else:
                field_type = field_info.annotation

            options_spec[key] = {
                'required': field_info.is_required(),
                'type': field_type,
                'is_flag': field_type is bool,
                'prompt': field_info.title,
                'default': default,
                'help': field_info.description,
            }
            for metadata in field_info.metadata:
                for metadata_key, metadata_value in metadata.items():
                    options_spec[key][metadata_key] = metadata_value

        options_ordered = []

        for name, spec in sorted(options_spec.items(), key=lambda x: x[1].get('priority', 0), reverse=True):
            spec.pop('priority', None)
            options_ordered.append(self.create_option(name, spec))

        return options_ordered

    @staticmethod
    def create_option(name, spec: dict) -> t.Callable[[t.Any], t.Any]:
        """Create a click option from a name and a specification."""
        is_flag = spec.pop('is_flag', False)
        name_dashed = name.replace('_', '-')
        option_name = f'--{name_dashed}/--no-{name_dashed}' if is_flag else f'--{name_dashed}'
        option_short_name = spec.pop('short_name', None)
        option_names = (option_short_name, option_name) if option_short_name else (option_name,)

        kwargs = {'cls': spec.pop('option_cls', InteractiveOption), 'show_default': True, 'is_flag': is_flag, **spec}

        # If the option is a flag with no default, make sure it is not prompted for, as that will force the user to
        # specify it to be on or off, but cannot let it unspecified.
        if kwargs['cls'] is InteractiveOption and is_flag and spec.get('default') is None:
            kwargs['cls'] = functools.partial(InteractiveOption, prompt_fn=lambda ctx: False)

        return click.option(*(option_names), **kwargs)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# ruff: noqa: E402
"""Modules related to the configuration of an AiiDA instance."""

from __future__ import annotations

# AUTO-GENERATED
# fmt: off
from .migrations import *
from .options import *
from .profile import *

__all__ = (
    'CURRENT_CONFIG_VERSION',
    'MIGRATIONS',
    'OLDEST_COMPATIBLE_CONFIG_VERSION',
    'Option',
    'Profile',
    'check_and_migrate_config',
    'config_needs_migrating',
    'downgrade_config',
    'get_current_version',
    'get_option',
    'get_option_names',
    'parse_option',
    'upgrade_config',
)

# fmt: on

# END AUTO-GENERATED


__all__ += (
    'get_config',
    'get_config_option',
    'get_config_path',
    'get_profile',
    'load_profile',
    'reset_config',
    'CONFIG',
)

import os
import warnings
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Optional

from aiida.common.warnings import AiidaDeprecationWarning

if TYPE_CHECKING:
    from aiida.orm import User

    from .config import Config

# global variables for aiida
CONFIG: Optional['Config'] = None


def get_config_path():
    """Returns path to .aiida configuration directory."""
    from .settings import AIIDA_CONFIG_FOLDER, DEFAULT_CONFIG_FILE_NAME

    return os.path.join(AIIDA_CONFIG_FOLDER, DEFAULT_CONFIG_FILE_NAME)


def load_config(create=False) -> 'Config':
    """Instantiate Config object representing an AiiDA configuration file.

    Warning: Contrary to :func:`~aiida.manage.configuration.get_config`, this function is uncached and will always
    create a new Config object. You may want to call :func:`~aiida.manage.configuration.get_config` instead.

    :param create: if True, will create the configuration file if it does not already exist
    :type create: bool

    :return: the config
    :rtype: :class:`~aiida.manage.configuration.config.Config`
    :raises aiida.common.MissingConfigurationError: if the configuration file could not be found and create=False
    """
    from aiida.common import exceptions

    from .config import Config

    filepath = get_config_path()

    if not os.path.isfile(filepath) and not create:
        raise exceptions.MissingConfigurationError(f'configuration file {filepath} does not exist')

    try:
        config = Config.from_file(filepath)
    except ValueError as exc:
        raise exceptions.ConfigurationError(f'configuration file {filepath} contains invalid JSON') from exc

    _merge_deprecated_cache_yaml(config, filepath)

    return config


def _merge_deprecated_cache_yaml(config, filepath):
    """Merge the deprecated cache_config.yml into the config."""
    cache_path = os.path.join(os.path.dirname(filepath), 'cache_config.yml')
    if not os.path.exists(cache_path):
        return

    # Imports are here to avoid them when not needed
    import shutil

    import yaml

    from aiida.common import timezone

    cache_path_backup = None
    # Keep generating a new backup filename based on the current time until it does not exist
    while not cache_path_backup or os.path.isfile(cache_path_backup):
        cache_path_backup = f"{cache_path}.{timezone.now().strftime('%Y%m%d-%H%M%S.%f')}"

    warnings.warn(
        'cache_config.yml use is deprecated and support will be removed in `v3.0`. Merging into config.json and '
        f'moving to: {cache_path_backup}',
        AiidaDeprecationWarning,
        stacklevel=2,
    )

    with open(cache_path, 'r', encoding='utf8') as handle:
        cache_config = yaml.safe_load(handle)
    for profile_name, data in cache_config.items():
        if profile_name not in config.profile_names:
            warnings.warn(f"Profile '{profile_name}' from cache_config.yml not in config.json, skipping", UserWarning)
            continue
        for key, option_name in [
            ('default', 'caching.default_enabled'),
            ('enabled', 'caching.enabled_for'),
            ('disabled', 'caching.disabled_for'),
        ]:
            if key in data:
                value = data[key]
                # in case of empty key
                value = [] if value is None and key != 'default' else value
                config.set_option(option_name, value, scope=profile_name)
    config.store()
    shutil.move(cache_path, cache_path_backup)


def load_profile(profile: Optional[str] = None, allow_switch=False) -> 'Profile':
    """Load a global profile, unloading any previously loaded profile.

    .. note:: if a profile is already loaded and no explicit profile is specified, nothing will be done

    :param profile: the name of the profile to load, by default will use the one marked as default in the config
    :param allow_switch: if True, will allow switching to a different profile when storage is already loaded

    :return: the loaded `Profile` instance
    :raises `aiida.common.exceptions.InvalidOperation`:
        if another profile has already been loaded and allow_switch is False
    """
    from aiida.manage import get_manager

    return get_manager().load_profile(profile, allow_switch)


def get_profile() -> Optional['Profile']:
    """Return the currently loaded profile.

    :return: the globally loaded `Profile` instance or `None`
    """
    from aiida.manage import get_manager

    return get_manager().get_profile()


@contextmanager
def profile_context(profile: 'Profile' | str | None = None, allow_switch=False) -> 'Profile':
    """Return a context manager for temporarily loading a profile, and unloading on exit.

    :param profile: the name of the profile to load, by default will use the one marked as default in the config
    :param allow_switch: if True, will allow switching to a different profile

    :return: a context manager for temporarily loading a profile
    """
    from aiida.manage import get_manager

    manager = get_manager()
    current_profile = manager.get_profile()
    yield manager.load_profile(profile, allow_switch)
    if current_profile is None:
        manager.unload_profile()
    else:
        manager.load_profile(current_profile, allow_switch=True)


def create_default_user(
    profile: Profile,
    email: str,
    first_name: Optional[str] = None,
    last_name: Optional[str] = None,
    institution: Optional[str] = None,
) -> User:
    """Create a default user for the given profile.

    If the profile's storage is read only, a random existing user will be queried and set as default. Otherwise a new
    user is created with the provided details and set as user.

    :param profile: The profile to create the user in.
    :param email: Email for the default user.
    :param first_name: First name for the default user.
    :param last_name: Last name for the default user.
    :param institution: Institution for the default user.
    :returns: The user that was set as the default user.
    """
    from aiida.manage import get_manager
    from aiida.orm import User

    with profile_context(profile, allow_switch=True):
        manager = get_manager()
        storage = manager.get_profile_storage()

        if storage.read_only:
            # Check if the storage contains any users, and just set a random one as default user.
            user = User.collection.query().first(flat=True)
        else:
            # Otherwise create a user and store it
            user = User(email=email, first_name=first_name, last_name=last_name, institution=institution).store()

        # The user can be ``None`` if the storage is read-only and doesn't contain any users. This shouldn't happen in
        # real situations, but this safe guard is added to be safe.
        if user:
            manager.set_default_user_email(profile, user.email)

    return user


def create_profile(
    config: 'Config',
    *,
    storage_backend: str,
    storage_config: dict[str, Any],
    broker_backend: str | None = None,
    broker_config: dict[str, Any] | None = None,
    name: str,
    email: str,
    first_name: Optional[str] = None,
    last_name: Optional[str] = None,
    institution: Optional[str] = None,
    is_test_profile: bool = False,
) -> Profile:
    """Create a new profile, initialise its storage and create a default user.

    :param config: The config instance.
    :param name: Name of the profile.
    :param email: Email for the default user.
    :param first_name: First name for the default user.
    :param last_name: Last name for the default user.
    :param institution: Institution for the default user.
    :param create_user: If `True`, creates a user that is set as the default user.
    :param storage_backend: The entry point to the :class:`aiida.orm.implementation.storage_backend.StorageBackend`
        implementation to use for the storage.
    :param storage_config: The configuration necessary to initialise and connect to the storage backend.
    :param broker_backend: The entry point to the :class:`aiida.brokers.Broker` implementation to use for the broker.
    :param broker_config: The configuration necessary to initialise and connect to the broker.
    """
    profile: Profile = config.create_profile(
        name=name,
        storage_backend=storage_backend,
        storage_config=storage_config,
        broker_backend=broker_backend,
        broker_config=broker_config,
        is_test_profile=is_test_profile,
    )

    create_default_user(profile, email, first_name, last_name, institution)

    return profile


def reset_config():
    """Reset the globally loaded config.

    .. warning:: This is experimental functionality and should for now be used only internally. If the reset is unclean
        weird unknown side-effects may occur that end up corrupting or destroying data.
    """
    global CONFIG  # noqa: PLW0603
    CONFIG = None


def get_config(create=False):
    """Return the current configuration.

    If the configuration has not been loaded yet
     * the configuration is loaded using ``load_config``
     * the global `CONFIG` variable is set
     * the configuration object is returned

    Note: This function will except if no configuration file can be found. Only call this function, if you need
    information from the configuration file.

    :param create: if True, will create the configuration file if it does not already exist
    :type create: bool

    :return: the config
    :rtype: :class:`~aiida.manage.configuration.config.Config`
    :raises aiida.common.ConfigurationError: if the configuration file could not be found, read or deserialized
    """
    global CONFIG  # noqa: PLW0603

    if not CONFIG:
        CONFIG = load_config(create=create)

        if CONFIG.get_option('warnings.showdeprecations'):
            # If the user does not want to get AiiDA deprecation warnings, we disable them - this can be achieved with::
            #   verdi config warnings.showdeprecations False
            # Note that the AiidaDeprecationWarning does NOT inherit from DeprecationWarning
            warnings.simplefilter('default', AiidaDeprecationWarning)
            # This should default to 'once', i.e. once per different message
        else:
            warnings.simplefilter('ignore', AiidaDeprecationWarning)

    return CONFIG


def get_config_option(option_name: str) -> Any:
    """Return the value of a configuration option.

    In order of priority, the option is returned from:

    1. The current profile, if loaded and the option specified
    2. The current configuration, if loaded and the option specified
    3. The default value for the option

    :param option_name: the name of the option to return
    :return: the value of the option
    :raises `aiida.common.exceptions.ConfigurationError`: if the option is not found
    """
    from aiida.manage import get_manager

    return get_manager().get_option(option_name)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The file format implementation"""

from pathlib import Path
from typing import Any, Literal, Union, overload

from aiida.storage.sqlite_zip.migrator import get_schema_version_head, migrate
from aiida.storage.sqlite_zip.utils import read_version
from aiida.tools.archive.abstract import ArchiveFormatAbstract

from .reader import ArchiveReaderSqlZip
from .writer import ArchiveAppenderSqlZip, ArchiveWriterSqlZip

__all__ = ('ArchiveFormatSqlZip',)


class ArchiveFormatSqlZip(ArchiveFormatAbstract):
    """Archive format, which uses a zip file, containing an SQLite database.

    The content of the zip file is::

        |- archive.zip
            |- metadata.json
            |- db.sqlite3
            |- repo/
                |- hashkey

    Repository files are named by their SHA256 content hash.

    """

    @property
    def latest_version(self) -> str:
        return get_schema_version_head()

    def read_version(self, path: Union[str, Path]) -> str:
        return read_version(path)

    @property
    def key_format(self) -> str:
        return 'sha256'

    @overload
    def open(
        self, path: Union[str, Path], mode: Literal['r'], *, compression: int = 6, **kwargs: Any
    ) -> ArchiveReaderSqlZip: ...

    @overload
    def open(
        self, path: Union[str, Path], mode: Literal['x', 'w'], *, compression: int = 6, **kwargs: Any
    ) -> ArchiveWriterSqlZip: ...

    @overload
    def open(
        self, path: Union[str, Path], mode: Literal['a'], *, compression: int = 6, **kwargs: Any
    ) -> ArchiveAppenderSqlZip: ...

    def open(
        self, path: Union[str, Path], mode: Literal['r', 'x', 'w', 'a'] = 'r', *, compression: int = 6, **kwargs: Any
    ) -> Union[ArchiveReaderSqlZip, ArchiveWriterSqlZip, ArchiveAppenderSqlZip]:
        if mode == 'r':
            return ArchiveReaderSqlZip(path, **kwargs)
        if mode == 'a':
            return ArchiveAppenderSqlZip(path, self, mode=mode, compression=compression, **kwargs)
        return ArchiveWriterSqlZip(path, self, mode=mode, compression=compression, **kwargs)

    def migrate(
        self,
        inpath: Union[str, Path],
        outpath: Union[str, Path],
        version: str,
        *,
        force: bool = False,
        compression: int = 6,
    ) -> None:
        """Migrate an archive to a specific version.

        :param path: archive path
        """
        return migrate(inpath, outpath, version, force=force, compression=compression)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module which provides decorators for AiiDA ORM entity -> DB field mappings."""

import datetime
import typing as t
from abc import ABCMeta
from copy import deepcopy
from functools import singledispatchmethod
from pprint import pformat

from aiida.common.lang import isidentifier

__all__ = (
    'QbField',
    'QbFieldFilters',
    'QbFields',
)


def extract_root_type(dtype: t.Any) -> t.Any:
    """Recursively search for the primitive root type.

    >>> extract_root_type(List[str]) -> list
    >>> extract_root_type(Optional[List[str]]) -> list
    """
    origin = t.get_origin(dtype)
    if origin:
        if origin is t.Union:
            return extract_root_type(t.get_args(dtype)[0])
        else:
            return origin
    else:
        return dtype


class QbField:
    """A field of an ORM entity, accessible via the ``QueryBuilder``"""

    __slots__ = (
        '_key',
        '_backend_key',
        '_doc',
        '_dtype',
        '_is_attribute',
        '_is_subscriptable',
    )

    def __init__(
        self,
        key: str,
        alias: t.Optional[str] = None,
        *,
        dtype: t.Optional[t.Any] = None,
        doc: str = '',
        is_attribute: bool = True,
        is_subscriptable: bool = False,
    ) -> None:
        """Initialise a ORM entity field, accessible via the ``QueryBuilder``

        :param key: The key of the field on the ORM entity
        :param alias: The alias in the storage backend for the key, if not equal to ``key``
        :param dtype: The data type of the field. If None, the field is of variable type.
        :param doc: A docstring for the field
        :param is_attribute: If True, the ``backend_key`` property will prepend "attributes." to field name
        :param is_subscriptable: If True, a new field can be created by ``field["subkey"]``
        """
        self._key = key
        self._backend_key = alias if alias is not None else key
        self._doc = doc
        self._dtype = dtype
        self._is_attribute = is_attribute
        self._is_subscriptable = is_subscriptable

    @property
    def key(self) -> str:
        return self._key

    @property
    def backend_key(self) -> str:
        if self._is_attribute:
            return f'attributes.{self._backend_key}'
        return self._backend_key

    @property
    def doc(self) -> str:
        return self._doc

    @property
    def dtype(self) -> t.Optional[t.Any]:
        """Return the primitive root type."""
        return extract_root_type(self._dtype)

    @property
    def annotation(self) -> t.Optional[t.Any]:
        """Return the full type annotation."""
        return self._dtype

    @property
    def is_attribute(self) -> bool:
        return self._is_attribute

    @property
    def is_subscriptable(self) -> bool:
        return self._is_subscriptable

    def __repr__(self) -> str:
        type_str = self._get_dtype_as_str() if self._dtype else ''
        return (
            f'{self.__class__.__name__}({self.key!r}'
            + (f', {self._backend_key!r}' if self._backend_key != self.key else '')
            + (f', dtype={type_str}' if self._dtype else '')
            + (f', is_attribute={self.is_attribute}')
            + (f', is_subscriptable={self.is_subscriptable!r}' if self.is_subscriptable else '')
            + ')'
        )

    def __str__(self) -> str:
        class_name = self.__class__.__name__
        field_name = f'{self.backend_key}{".*" if self.is_subscriptable else ""}'
        type_str = self._get_dtype_as_str()
        return f'{class_name}({field_name}) -> {type_str}'

    def __hash__(self):
        return hash((self.key, self.backend_key))

    def __eq__(self, value):
        return QbFieldFilters(((self, '==', value),))

    def __ne__(self, value):
        return QbFieldFilters(((self, '!==', value),))

    def in_(self, value: t.Iterable[t.Any]):
        """Return a filter for only values in the list"""
        try:
            set(value)
        except TypeError:
            raise TypeError('in_ must be iterable')
        return QbFieldFilters(((self, 'in', value),))

    def _get_dtype_as_str(self) -> str:
        """Return field type as processed string.

        >>> None -> ?
        >>> str -> str
        >>> typing.Optional[str] -> Optional[str]
        >>> typing.Dict[typing.List[str]] -> Dict[List[str]]
        """
        if self._dtype:
            if t.get_origin(self._dtype):
                type_str = str(self._dtype).replace('typing.', '')
            else:
                type_str = self._dtype.__name__
        else:
            type_str = '?'
        return type_str

    if t.TYPE_CHECKING:

        def __getitem__(self, key: str) -> 'QbField': ...


class QbNumericField(QbField):
    """A numeric (`int`, `float`, `datetime`) flavor of `QbField`."""

    def __lt__(self, value):
        return QbFieldFilters(((self, '<', value),))

    def __le__(self, value):
        return QbFieldFilters(((self, '<=', value),))

    def __gt__(self, value):
        return QbFieldFilters(((self, '>', value),))

    def __ge__(self, value):
        return QbFieldFilters(((self, '>=', value),))


class QbArrayField(QbField):
    """An array (`list`) flavor of `QbField`."""

    def contains(self, value):
        """Return a filter for only values containing these items"""
        return QbFieldFilters(((self, 'contains', value),))

    def of_length(self, value: int):
        """Return a filter for only array values of this length."""
        if not isinstance(value, int):
            raise TypeError('of_length must be an integer')
        return QbFieldFilters(((self, 'of_length', value),))

    def longer(self, value: int):
        """Return a filter for only array values longer than this length."""
        if not isinstance(value, int):
            raise TypeError('longer must be an integer')
        return QbFieldFilters(((self, 'longer', value),))

    def shorter(self, value: int):
        """Return a filter for only array values shorter than this length."""
        if not isinstance(value, int):
            raise TypeError('shorter must be an integer')
        return QbFieldFilters(((self, 'shorter', value),))


class QbStrField(QbField):
    """A string (`str`) flavor of `QbField`."""

    def like(self, value: str):
        """Return a filter for only string values matching the wildcard string.

        - The percent sign (`%`) represents zero, one, or multiple characters
        - The underscore sign (`_`) represents one, single character
        """
        if not isinstance(value, str):
            raise TypeError('like must be a string')
        return QbFieldFilters(((self, 'like', value),))

    def ilike(self, value: str):
        """Return a filter for only string values matching the (case-insensitive) wildcard string.

        - The percent sign (`%`) represents zero, one, or multiple characters
        - The underscore sign (`_`) represents one, single character
        """
        if not isinstance(value, str):
            raise TypeError('ilike must be a string')
        return QbFieldFilters(((self, 'ilike', value),))


class QbDictField(QbField):
    """A dictionary (`dict`) flavor of `QbField`."""

    def has_key(self, value):
        """Return a filter for only values with these keys"""
        return QbFieldFilters(((self, 'has_key', value),))

    def __getitem__(self, key: str) -> 'QbAttrField':
        """Return a new `QbField` with a nested key."""
        if not self.is_subscriptable:
            raise IndexError('This field is not subscriptable')
        return QbAttrField(
            f'{self.key}.{key}',
            f'{self._backend_key}.{key}' if self.is_attribute else None,
            is_attribute=self.is_attribute,
        )


class QbAttrField(QbNumericField, QbArrayField, QbStrField, QbDictField):
    """A generic flavor of `QbField` covering all operations."""

    def of_type(self, value):
        """Return a filter for only values of this type."""
        return QbFieldFilters(((self, 'of_type', value),))


class QbFieldFilters:
    """An representation of a list of fields and their comparators."""

    __slots__ = ('filters',)

    def __init__(
        self,
        filters: t.Union[t.Sequence[t.Tuple[QbField, str, t.Any]], dict],
    ):
        self.filters: t.Dict[str, t.Any] = {}
        self.add_filters(filters)

    def as_dict(self) -> t.Dict[str, t.Any]:
        """Return the filters dictionary."""
        return self.filters

    def items(self):
        """Return an items view of the filters for use in the QueryBuilder."""
        return self.filters.items()

    @singledispatchmethod
    def add_filters(self, filters: dict):
        self.filters.update(filters)

    @add_filters.register(list)
    @add_filters.register(tuple)
    def _(self, filters):
        field: QbField
        for field, comparator, value in filters:
            qb_field = field.backend_key
            if qb_field in self.filters:
                self.filters['and'] = [
                    {qb_field: self.filters.pop(qb_field)},
                    {qb_field: {comparator: value}},
                ]
            else:
                self.filters[qb_field] = {comparator: value}

    def __repr__(self) -> str:
        return f'QbFieldFilters({self.filters})'

    def __getitem__(self, key: str) -> t.Any:
        return self.filters[key]

    def __contains__(self, key: str) -> bool:
        return key in self.filters

    def __eq__(self, other: object) -> bool:
        """``a == b`` checks if `a.filters == b.filters`."""
        if not isinstance(other, QbFieldFilters):
            raise TypeError(f'Cannot compare QbFieldFilters to {type(other)}')
        return self.filters == other.filters

    def __and__(self, other: 'QbFieldFilters') -> 'QbFieldFilters':
        """``a & b`` -> {'and': [`a.filters`, `b.filters`]}."""
        return self._resolve_redundancy(other, 'and') or QbFieldFilters({'and': [self.filters, other.filters]})

    def __or__(self, other: 'QbFieldFilters') -> 'QbFieldFilters':
        """``a | b`` -> {'or': [`a.filters`, `b.filters`]}."""
        return self._resolve_redundancy(other, 'or') or QbFieldFilters({'or': [self.filters, other.filters]})

    def __invert__(self) -> 'QbFieldFilters':
        """~(a > b) -> a !> b; ~(a !> b) -> a > b"""
        filters = deepcopy(self.filters)
        if 'and' in filters:
            filters['!and'] = filters.pop('and')
        elif 'or' in filters:
            filters['!or'] = filters.pop('or')
        elif '!and' in filters:
            filters['and'] = filters.pop('!and')
        elif '!or' in filters:
            filters['or'] = filters.pop('!or')
        else:
            key, args = next(iter(filters.items()))
            operator, value = next(iter(args.items()))
            operator = operator[1:] if '!' in operator else f'!{operator}'
            filters[key] = {operator: value}
        return QbFieldFilters(filters)

    def _resolve_redundancy(self, other: 'QbFieldFilters', logical: str) -> t.Optional['QbFieldFilters']:
        """Resolve redundant filters and nested logical operators."""

        if not isinstance(other, QbFieldFilters):
            raise TypeError(f'Cannot combine QbFieldFilters and {type(other)}')

        # same filters
        if other == self:
            return self

        # self is already wrapped in `logical`
        # append other to self
        if logical in self.filters:
            self[logical].append(other.filters)
            return self

        # other is already wrapped in `logical`
        # insert self in other
        if logical in other:
            other[logical].insert(0, self.filters)
            return other

        return None


class QbFields:
    """A readonly class for mapping attributes to database fields of an AiiDA entity."""

    __isabstractmethod__ = False

    def __init__(self, fields: t.Optional[t.Dict[str, QbField]] = None):
        self._fields = fields or {}

    def __repr__(self) -> str:
        return pformat({key: str(value) for key, value in self._fields.items()})

    def __str__(self) -> str:
        return str({key: str(value) for key, value in self._fields.items()})

    def __getitem__(self, key: str) -> QbField:
        """Return an QbField by key."""
        return self._fields[key]

    def __getattr__(self, key: str) -> QbField:
        """Return an QbField by key."""
        try:
            return self._fields[key]
        except KeyError:
            raise AttributeError(key)

    def __contains__(self, key: str) -> bool:
        """Return if the field key exists"""
        return key in self._fields

    def __len__(self) -> int:
        """Return the number of fields"""
        return len(self._fields)

    def __iter__(self):
        """Iterate through the field keys"""
        return iter(self._fields)

    def __dir__(self):
        """Return keys for tab competion."""
        return list(self._fields) + ['_dict']

    @property
    def _dict(self):
        """Return a copy of the internal mapping"""
        return deepcopy(self._fields)


class EntityFieldMeta(ABCMeta):
    """A metaclass for entity fields, which adds a `fields` class attribute."""

    def __init__(cls, name, bases, classdict):
        super().__init__(name, bases, classdict)

        # only allow an existing fields attribute if has been generated from a subclass
        current_fields = getattr(cls, 'fields', None)
        if current_fields is not None and not isinstance(current_fields, QbFields):
            raise ValueError(f"class '{cls}' already has a `fields` attribute set")

        # Find all fields
        fields = {}
        # Note: inspect.getmembers causes loading of AiiDA to fail
        for key, attr in ((key, attr) for subcls in reversed(cls.__mro__) for key, attr in subcls.__dict__.items()):
            # __qb_fields__ should be a list of QbField instances
            if key == '__qb_fields__':
                assert isinstance(
                    attr, t.Sequence
                ), f"class '{cls}' has a '__qb_fields__' attribute, but it is not a sequence"
                for field in attr:
                    if not isinstance(field, QbField):
                        raise ValueError(f"__qb_fields__ attribute of class '{cls}' must be list of QbField instances")
                    if field.key == 'value':
                        if dtype := getattr(cls, '_type', None):
                            value_field = add_field(
                                'value',
                                dtype=dtype,
                                doc='The value of the data',
                            )

                            fields[field.key] = value_field
                    else:
                        fields[field.key] = field

        cls.fields = QbFields({key: fields[key] for key in sorted(fields)})


class QbFieldArguments(t.TypedDict):
    key: str
    alias: t.Optional[str]
    dtype: t.Optional[t.Any]
    doc: str
    is_attribute: bool
    is_subscriptable: bool


def add_field(
    key: str,
    alias: t.Optional[str] = None,
    *,
    dtype: t.Optional[t.Any] = None,
    doc: str = '',
    is_attribute: bool = True,
    is_subscriptable: bool = False,
) -> QbField:
    """Add a `dtype`-dependent `QbField` representation of a field.

    :param key: The key of the field on the ORM entity
    :param alias: The alias in the storage backend for the key, if not equal to ``key``
    :param dtype: The data type of the field. If None, the field is of variable type.
    :param doc: A docstring for the field
    :param is_attribute: If True, the ``backend_key`` property will prepend "attributes." to field name
    :param is_subscriptable: If True, a new field can be created by ``field["subkey"]``
    """
    kwargs: QbFieldArguments = {
        'key': key,
        'alias': alias,
        'dtype': dtype,
        'doc': doc,
        'is_attribute': is_attribute,
        'is_subscriptable': is_subscriptable,
    }
    if not isidentifier(key):
        raise ValueError(f'{key} is not a valid python identifier')
    if not is_attribute and alias:
        raise ValueError('only attribute fields may be aliased')
    if not dtype:
        return QbField(**kwargs)
    root_type = extract_root_type(dtype)
    if root_type in (int, float, datetime.datetime):
        return QbNumericField(**kwargs)
    elif root_type is list:
        return QbArrayField(**kwargs)
    elif root_type is str:
        return QbStrField(**kwargs)
    elif root_type is dict:
        return QbDictField(**kwargs)
    else:
        return QbField(**kwargs)



#!/usr/bin/env python
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility CLI to manage dependencies for aiida-core."""

import os
import re
import subprocess
import sys
from collections import OrderedDict, defaultdict
from pathlib import Path

import click
import requests
import tomli
import yaml
from packaging.requirements import Requirement
from packaging.specifiers import Specifier
from packaging.utils import canonicalize_name
from packaging.version import parse

ROOT = Path(__file__).resolve().parent.parent  # repository root

SETUPTOOLS_CONDA_MAPPINGS = {
    'graphviz': 'python-graphviz',
    'docstring-parser': 'docstring_parser',
}

CONDA_IGNORE = []

GITHUB_ACTIONS = os.environ.get('GITHUB_ACTIONS') == 'true'


class DependencySpecificationError(click.ClickException):
    """Indicates an issue in a dependency specification."""


def _load_pyproject():
    """Load the setup configuration from the 'pyproject.toml' file."""
    try:
        with open(ROOT / 'pyproject.toml', 'rb') as handle:
            return tomli.load(handle)
    except tomli.TOMLDecodeError as error:
        raise DependencySpecificationError(f"Error while parsing 'pyproject.toml' file: {error}")
    except FileNotFoundError:
        raise DependencySpecificationError("The 'pyproject.toml' file is missing!")


def _load_environment_yml():
    """Load the conda environment specification from the 'environment.yml' file."""
    try:
        with open(ROOT / 'environment.yml', encoding='utf8') as file:
            return yaml.load(file, Loader=yaml.SafeLoader)
    except yaml.error.YAMLError as error:
        raise DependencySpecificationError(f"Error while parsing 'environment.yml':\n{error}")
    except FileNotFoundError as error:
        raise DependencySpecificationError(str(error))


def _setuptools_to_conda(req):
    """Map package names from setuptools to conda where necessary.

    In case that the same underlying dependency is listed under different names
    on PyPI and conda-forge.
    """
    for pattern, replacement in SETUPTOOLS_CONDA_MAPPINGS.items():
        if re.match(pattern, str(req)):
            req = Requirement(re.sub(pattern, replacement, str(req)))
            break

    # markers are not supported by conda
    req.marker = None

    # We need to parse the modified required again, to ensure consistency.
    return Requirement(str(req))


def _find_linenos_of_requirements_in_pyproject(requirements):
    """Determine the line numbers of requirements specified in 'pyproject.toml'.

    Returns a dict that maps a requirement, e.g., `numpy~=1.15.0` to the
    line numbers at which said requirement is defined within the 'pyproject.toml'
    file.
    """
    linenos = defaultdict(list)

    with open(ROOT / 'pyproject.toml', encoding='utf8') as setup_json_file:
        lines = list(setup_json_file)

    # Determine the lines that correspond to affected requirements in pyproject.toml.
    for requirement in requirements:
        for lineno, line in enumerate(lines):
            if str(requirement) in line:
                linenos[requirement].append(lineno)
    return linenos


def parse_requirements(requirements):
    """Parse requirements from a file or list of strings."""
    results = []
    for requirement in requirements:
        stripped = requirement.strip()
        if stripped and not stripped.startswith('#'):
            results.append(Requirement(stripped))
    return results


@click.group()
def cli():
    """Manage dependencies of the aiida-core package."""


@cli.command('generate-environment-yml')
def generate_environment_yml():
    """Generate 'environment.yml' file."""
    # needed for ordered dict, see https://stackoverflow.com/a/52621703
    yaml.add_representer(
        OrderedDict,
        lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()),
        Dumper=yaml.SafeDumper,
    )

    # Read the requirements from 'pyproject.toml'
    pyproject = _load_pyproject()
    install_requirements = [Requirement(r) for r in pyproject['project']['dependencies']]

    # python version cannot be overriden from outside environment.yml
    # (even if it is not specified at all in environment.yml)
    # https://github.com/conda/conda/issues/9506
    conda_requires = ['python~=3.9']
    for req in install_requirements:
        if req.name == 'python' or any(re.match(ignore, str(req)) for ignore in CONDA_IGNORE):
            continue
        conda_requires.append(str(_setuptools_to_conda(req)))

    environment = OrderedDict(
        [
            ('name', 'aiida'),
            ('channels', ['conda-forge', 'defaults']),
            ('dependencies', conda_requires),
        ]
    )

    with open(ROOT / 'environment.yml', 'w', encoding='utf8') as env_file:
        env_file.write('# Usage: conda env create -n myenvname -f environment.yml\n')
        yaml.safe_dump(
            environment, env_file, explicit_start=True, default_flow_style=False, encoding='utf-8', allow_unicode=True
        )


@cli.command()
@click.pass_context
def generate_all(ctx):
    """Generate all dependent requirement files."""
    ctx.invoke(generate_environment_yml)


@cli.command('validate-environment-yml', help="Validate 'environment.yml'.")
def validate_environment_yml():
    """Validate that 'environment.yml' is consistent with 'pyproject.toml'."""
    # Read the requirements from 'pyproject.toml' and 'environment.yml'.
    pyproject = _load_pyproject()
    install_requirements = [Requirement(r) for r in pyproject['project']['dependencies']]
    python_requires = Requirement('python' + pyproject['project']['requires-python'])

    environment_yml = _load_environment_yml()
    try:
        assert environment_yml['name'] == 'aiida', "environment name should be 'aiida'."
        assert environment_yml['channels'] == [
            'conda-forge',
            'defaults',
        ], "channels should be 'conda-forge', 'defaults'."
    except AssertionError as error:
        raise DependencySpecificationError(f"Error in 'environment.yml': {error}")

    try:
        conda_dependencies = {Requirement(d) for d in environment_yml['dependencies']}
    except TypeError as error:
        raise DependencySpecificationError(f"Error while parsing requirements from 'environment_yml': {error}")

    # Attempt to find the specification of Python among the 'environment.yml' dependencies.
    for dependency in conda_dependencies:
        if dependency.name == 'python':  # Found the Python dependency specification
            conda_python_dependency = dependency
            conda_dependencies.remove(dependency)
            break
    else:  # Failed to find Python dependency specification
        raise DependencySpecificationError("Did not find specification of Python version in 'environment.yml'.")

    # The Python version specified in 'pyproject.toml' should be listed as trove classifiers.
    for spec in conda_python_dependency.specifier:
        expected_classifier = 'Programming Language :: Python :: ' + spec.version
        if expected_classifier not in pyproject['project']['classifiers']:
            raise DependencySpecificationError(
                f"Trove classifier '{expected_classifier}' missing from 'pyproject.toml'."
            )

        # The Python version should be specified as supported in 'pyproject.toml'.
        if not any(spec.version >= other_spec.version for other_spec in python_requires.specifier):
            raise DependencySpecificationError(
                f"Required Python version {spec.version} from 'environment.yaml' is not consistent with "
                + "required version in 'pyproject.toml'."
            )

        break
    else:
        raise DependencySpecificationError(f"Missing specifier: '{conda_python_dependency}'.")

    # Check that all requirements specified in the pyproject.toml file are found in the
    # conda environment specification.
    for req in install_requirements:
        if any(re.match(ignore, str(req)) for ignore in CONDA_IGNORE):
            continue  # skip explicitly ignored packages

        try:
            conda_dependencies.remove(_setuptools_to_conda(req))
        except KeyError:
            raise DependencySpecificationError(f"Requirement '{req}' not specified in 'environment.yml'.")

    # The only dependency left should be the one for Python itself, which is not part of
    # the install_requirements for setuptools.
    if conda_dependencies:
        raise DependencySpecificationError(
            "The 'environment.yml' file contains dependencies that are missing " "in 'pyproject.toml':\n- {}".format(
                '\n- '.join(map(str, conda_dependencies))
            )
        )

    click.secho('Conda dependency specification is consistent.', fg='green')


@cli.command('validate-all', help='Validate consistency of all requirements.')
@click.pass_context
def validate_all(ctx):
    """Validate consistency of all requirement specifications of the package.

    Validates that the specification of requirements/dependencies is consistent across
    the following files:

    - pyproject.toml
    - environment.yml

    """
    ctx.invoke(validate_environment_yml)


@cli.command()
@click.argument('extras', nargs=-1)
@click.option(
    '--github-annotate/--no-github-annotate',
    default=True,
    hidden=True,
    help='Control whether to annotate files with context-specific warnings '
    'as part of a GitHub actions workflow. Note: Requires environment '
    'variable GITHUB_ACTIONS=true .',
)
def check_requirements(extras, github_annotate):
    """Check the 'requirements/*.txt' files.

    Checks that the environments specified in the requirements files
    match all the dependencies specified in 'pyproject.toml'.

    The arguments allow to specify which 'extra' requirements to expect.
    Use 'DEFAULT' to select 'atomic_tools', 'docs', 'notebook', 'rest', and 'tests'.

    """
    if len(extras) == 1 and extras[0] == 'DEFAULT':
        extras = ['atomic_tools', 'docs', 'notebook', 'rest', 'tests']

    # Read the requirements from 'pyproject.toml''
    pyproject = _load_pyproject()
    install_requires = pyproject['project']['dependencies']
    for extra in extras:
        install_requires.extend(pyproject['project']['optional-dependencies'][extra])
    install_requires = set(parse_requirements(install_requires))

    not_installed = defaultdict(list)
    for fn_req in (ROOT / 'requirements').iterdir():
        match = re.match(r'.*-py-(.*)\.txt', str(fn_req))
        if not match:
            continue
        env = {'python_version': match.groups()[0]}
        requirements_abstract = {r for r in install_requires if r.marker is None or r.marker.evaluate(env)}
        installed = []

        with open(fn_req, encoding='utf8') as req_file:
            requirements_concrete = parse_requirements(req_file)

        for requirement_abstract in requirements_abstract:
            for requirement_concrete in requirements_concrete:
                version = Specifier(str(requirement_concrete.specifier)).version
                if canonicalize_name(requirement_abstract.name) == canonicalize_name(
                    requirement_concrete.name
                ) and requirement_abstract.specifier.contains(version):
                    installed.append(requirement_abstract)
                    break

        for dependency in requirements_abstract.difference(set(installed)):
            not_installed[dependency].append(fn_req)

    if any(not_installed.values()):
        setup_json_linenos = _find_linenos_of_requirements_in_pyproject(not_installed)

        # Format error message to be presented to user.
        error_msg = ["The requirements/ files are missing dependencies specified in the 'pyproject.toml' file.", '']
        for dependency, fn_reqs in not_installed.items():
            src = 'pyproject.toml' + ','.join(str(lineno + 1) for lineno in setup_json_linenos[dependency])
            error_msg.append(f'{src}: No match for dependency `{dependency}` in:')
            for fn_req in sorted(fn_reqs):
                error_msg.append(f' - {fn_req.relative_to(ROOT)}')

        if GITHUB_ACTIONS:
            # Set the step ouput error message which can be used, e.g., for display as part of an issue comment.
            print('::set-output name=error::' + '%0A'.join(error_msg))

        if GITHUB_ACTIONS and github_annotate:
            # Annotate the pyproject.toml' file with specific warnings.
            for dependency, fn_reqs in not_installed.items():
                for lineno in setup_json_linenos[dependency]:
                    print(
                        f'::warning file=pyproject.toml,line={lineno+1}::'
                        f"No match for dependency '{dependency}' in: "
                        + ','.join(str(fn_req.relative_to(ROOT)) for fn_req in fn_reqs)
                    )

        raise DependencySpecificationError('\n'.join(error_msg))

    click.secho("Requirements files appear to be in sync with specifications in 'pyproject.toml''.", fg='green')


@cli.command()
@click.argument('extras', nargs=-1)
@click.option('--format', 'fmt', type=click.Choice(['pip', 'pipfile']), default='pip')
def show_requirements(extras, fmt):
    """Show the installation requirements.

    For example:

        show-requirements --format=pipfile all

    This will show all reqiurements including *all* extras in Pipfile format.
    """
    # Read the requirements from 'pyproject.toml''
    pyproject = _load_pyproject()

    if 'all' in extras:
        extras = list(pyproject['project']['optional-dependencies'])

    to_install = {Requirement(r) for r in pyproject['project']['dependencies']}
    for key in extras:
        to_install.update(Requirement(r) for r in pyproject['project']['optional-dependencies'][key])

    if fmt == 'pip':
        click.echo('\n'.join(sorted(map(str, to_install))))
    elif fmt == 'pipfile':
        click.echo('[packages]')
        for requirement in sorted(to_install, key=str):
            click.echo(f'{requirement.name} = "{requirement.specifier}"')


@cli.command()
@click.argument('extras', nargs=-1)
def pip_install_extras(extras):
    """Install extra requirements.

    For example:

        pip-install-extras docs

    This will install *only* the extra the requirements for docs, but without triggering
    the installation of the main installations requirements of the aiida-core package.
    """
    # Read the requirements from 'pyproject.toml''
    pyproject = _load_pyproject()

    to_install = set()
    for key in extras:
        to_install.update(Requirement(r) for r in pyproject['project']['optional-dependencies'][key])

    cmd = [sys.executable, '-m', 'pip', 'install'] + [str(r) for r in to_install]
    subprocess.run(cmd, check=True)


@cli.command()
@click.argument('extras', nargs=-1)
@click.option('--pre-releases', is_flag=True, help='Include pre-releases.')
def identify_outdated(extras, pre_releases):
    """Identify outdated dependencies.

    For example:

        identify-outdated all

    This command will analyze the current dependencies and compare them against
    the latest versions released on PyPI. It then lists all dependencies where
    the latest release is not compatible with the dependency specification.
    This function can thus be used to identify dependencies where the
    specification must be loosened.
    """
    # Read the requirements from 'pyproject.toml''
    pyproject = _load_pyproject()

    to_install = {Requirement(r) for r in pyproject['project']['dependencies']}
    for key in extras:
        to_install.update(Requirement(r) for r in pyproject['project']['optional-dependencies'][key])

    def get_package_data(name):
        req = requests.get(f'https://pypi.python.org/pypi/{name}/json', timeout=5)
        req.raise_for_status()
        return req.json()

    release_data = {requirement: get_package_data(requirement.name)['releases'] for requirement in to_install}
    for requirement, releases in release_data.items():
        releases_ = list(sorted(map(parse, releases)))
        latest_release = [r for r in releases_ if pre_releases or not r.is_prerelease][-1]

        if str(latest_release) not in requirement.specifier:
            print(requirement, latest_release)


if __name__ == '__main__':
    cli()



"""Class that defines the abstract interface for an object repository.

The scope of this class is intentionally very narrow. Any backend implementation should merely provide the methods to
store binary blobs, or "objects", and return a string-based key that unique identifies the object that was just created.
This key should then be able to be used to retrieve the bytes of the corresponding object or to delete it.
"""

import abc
import contextlib
import hashlib
import io
import pathlib
from typing import BinaryIO, Iterable, Iterator, List, Optional, Tuple, Union

from aiida.common.hashing import chunked_file_hash

__all__ = ('AbstractRepositoryBackend',)


class AbstractRepositoryBackend(metaclass=abc.ABCMeta):
    """Class that defines the abstract interface for an object repository.

    The repository backend only deals with raw bytes, both when creating new objects as well as when returning a stream
    or the content of an existing object. The encoding and decoding of the byte content should be done by the client
    upstream. The file repository backend is also not expected to keep any kind of file hierarchy but must be assumed
    to be a simple flat data store. When files are created in the file object repository, the implementation will return
    a string-based key with which the content of the stored object can be addressed. This key is guaranteed to be unique
    and persistent. Persisting the key or mapping it onto a virtual file hierarchy is again up to the client upstream.
    """

    @property
    @abc.abstractmethod
    def uuid(self) -> Optional[str]:
        """Return the unique identifier of the repository."""

    @property
    @abc.abstractmethod
    def key_format(self) -> Optional[str]:
        """Return the format for the keys of the repository.

        Important for when migrating between backends (e.g. archive -> main), as if they are not equal then it is
        necessary to re-compute all the `Node.base.repository.metadata` before importing (otherwise they will not match
        with the repository).
        """

    @abc.abstractmethod
    def initialise(self, **kwargs) -> None:
        """Initialise the repository if it hasn't already been initialised.

        :param kwargs: parameters for the initialisation.
        """

    @property
    @abc.abstractmethod
    def is_initialised(self) -> bool:
        """Return whether the repository has been initialised."""

    @abc.abstractmethod
    def erase(self) -> None:
        """Delete the repository itself and all its contents.

        .. note:: This should not merely delete the contents of the repository but any resources it created. For
            example, if the repository is essentially a folder on disk, the folder itself should also be deleted, not
            just its contents.
        """

    @staticmethod
    def is_readable_byte_stream(handle) -> bool:
        return hasattr(handle, 'read') and hasattr(handle, 'mode') and 'b' in handle.mode

    def put_object_from_filelike(self, handle: BinaryIO) -> str:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :return: the generated fully qualified identifier for the object within the repository.
        :raises TypeError: if the handle is not a byte stream.
        """
        if (
            not isinstance(handle, io.BufferedIOBase)  # type: ignore[redundant-expr,unreachable]
            and not self.is_readable_byte_stream(handle)
        ):
            raise TypeError(f'handle does not seem to be a byte stream: {type(handle)}.')
        return self._put_object_from_filelike(handle)

    @abc.abstractmethod
    def _put_object_from_filelike(self, handle: BinaryIO) -> str:
        pass

    def put_object_from_file(self, filepath: Union[str, pathlib.Path]) -> str:
        """Store a new object with contents of the file located at `filepath` on this file system.

        :param filepath: absolute path of file whose contents to copy to the repository.
        :return: the generated fully qualified identifier for the object within the repository.
        :raises TypeError: if the handle is not a byte stream.
        """
        with open(filepath, mode='rb') as handle:
            return self.put_object_from_filelike(handle)

    @abc.abstractmethod
    def has_objects(self, keys: List[str]) -> List[bool]:
        """Return whether the repository has an object with the given key.

        :param keys:
            list of fully qualified identifiers for objects within the repository.
        :return:
            list of logicals, in the same order as the keys provided, with value True if the respective
            object exists and False otherwise.
        """

    def has_object(self, key: str) -> bool:
        """Return whether the repository has an object with the given key.

        :param key: fully qualified identifier for the object within the repository.
        :return: True if the object exists, False otherwise.
        """
        return self.has_objects([key])[0]

    @abc.abstractmethod
    def list_objects(self) -> Iterable[str]:
        """Return iterable that yields all available objects by key.

        :return: An iterable for all the available object keys.
        """

    @abc.abstractmethod
    def get_info(self, detailed: bool = False, **kwargs) -> dict:
        """Returns relevant information about the content of the repository.

        :param detailed:
            flag to enable extra information (detailed=False by default, only returns basic information).

        :return: a dictionary with the information.
        """

    @abc.abstractmethod
    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        """Performs maintenance operations.

        :param dry_run:
            flag to only print the actions that would be taken without actually executing them.

        :param live:
            flag to indicate to the backend whether AiiDA is live or not (i.e. if the profile of the
            backend is currently being used/accessed). The backend is expected then to only allow (and
            thus set by default) the operations that are safe to perform in this state.
        """

    @contextlib.contextmanager
    def open(self, key: str) -> Iterator[BinaryIO]:  # type: ignore[return]
        """Open a file handle to an object stored under the given key.

        .. note:: this should only be used to open a handle to read an existing file. To write a new file use the method
            ``put_object_from_filelike`` instead.

        :param key: fully qualified identifier for the object within the repository.
        :return: yield a byte stream object.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if the file could not be opened.
        """
        if not self.has_object(key):
            raise FileNotFoundError(f'object with key `{key}` does not exist.')

    def get_object_content(self, key: str) -> bytes:
        """Return the content of a object identified by key.

        :param key: fully qualified identifier for the object within the repository.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if the file could not be opened.
        """
        with self.open(key) as handle:
            return handle.read()

    @abc.abstractmethod
    def iter_object_streams(self, keys: List[str]) -> Iterator[Tuple[str, BinaryIO]]:
        """Return an iterator over the (read-only) byte streams of objects identified by key.

        .. note:: handles should only be read within the context of this iterator.

        :param keys: fully qualified identifiers for the objects within the repository.
        :return: an iterator over the object byte streams.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if a file could not be opened.
        """

    def get_object_hash(self, key: str) -> str:
        """Return the SHA-256 hash of an object stored under the given key.

        .. important::
            A SHA-256 hash should always be returned,
            to ensure consistency across different repository implementations.

        :param key: fully qualified identifier for the object within the repository.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if the file could not be opened.
        """
        with self.open(key) as handle:
            return chunked_file_hash(handle, hashlib.sha256)

    @abc.abstractmethod
    def delete_objects(self, keys: List[str]) -> None:
        """Delete the objects from the repository.

        :param keys: list of fully qualified identifiers for the objects within the repository.
        :raise FileNotFoundError: if any of the files does not exist.
        :raise OSError: if any of the files could not be deleted.
        """
        keys_exist = self.has_objects(keys)
        if not all(keys_exist):
            error_message = 'some of the keys provided do not correspond to any object in the repository:\n'
            for indx, key_exists in enumerate(keys_exist):
                if not key_exists:
                    error_message += f' > object with key `{keys[indx]}` does not exist.\n'
            raise FileNotFoundError(error_message)

    def delete_object(self, key: str) -> None:
        """Delete the object from the repository.

        :param key: fully qualified identifier for the object within the repository.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if the file could not be deleted.
        """
        return self.delete_objects([key])



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Miscellaneous functions for escaping strings."""

import re


def escape_for_bash(str_to_escape, use_double_quotes=False):
    """This function takes any string and escapes it in a way that
    bash will interpret it as a single string.

    Explanation:

    At the end, in the return statement, the string is put within single
    quotes. Therefore, the only thing that I have to escape in bash is the
    single quote character. To do this, I substitute every single
    quote ' with '"'"' which means:

    First single quote: exit from the enclosing single quotes

    Second, third and fourth character: "'" is a single quote character,
    escaped by double quotes

    Last single quote: reopen the single quote to continue the string

    Finally, note that for python I have to enclose the string '"'"'
    within triple quotes to make it work, getting finally: the complicated
    string found below.

    :param str_to_escape: the string to escape.
    :param use_double_quotes: boolean, if ``True``, use double quotes instead of single quotes.
    :return: the escaped string.
    """
    if str_to_escape is None:
        return ''

    str_to_escape = str(str_to_escape)
    if use_double_quotes:
        escaped_quotes = str_to_escape.replace('"', '''"'"'"''')
        escaped = f'"{escaped_quotes}"'
    else:
        escaped_quotes = str_to_escape.replace("'", """'"'"'""")
        escaped = f"'{escaped_quotes}'"

    return escaped


# Mapping of "SQL" tokens into corresponding regex expressions
SQL_TO_REGEX_TOKENS = [  # Remember in the strings below we have to escape backslashes as well for python...
    # so '\\\\' is actually a string with two backslashes, '\\' a string with a single backslash, ...
    ('\\\\', re.escape('\\')),  # Double slash should be interpreted as a literal single backslash by regex
    ('\\%', re.escape('%')),  # literal '\%' should be interpreted as literal % by regex
    ('\\_', re.escape('_')),  # literal '\_' should be interpreted as literal _ by regex
    ('%', '.*'),
    ('_', '.'),
]


def escape_for_sql_like(string):
    """Function that escapes % or _ symbols provided by user

    SQL LIKE syntax summary:

    - ``%`` -> match any number of characters
    - ``_`` -> match exactly one character

    """
    return string.replace('%', '\\%').replace('_', '\\_')


def get_regex_pattern_from_sql(sql_pattern):
    r"""Convert a string providing a pattern to match in SQL
    syntax into a string performing the same match as a regex.

    SQL LIKE syntax summary:

    - ``%`` -> match any number of characters
    - ``_`` -> match exactly one character

    Moreover, ``\`` is the escape character (by default), so:

    - ``\\`` -> single backslash
    - ``\%`` -> literal % symbol
    - ``\_`` -> literal _ symbol

    and moreover the string should begin at the beginning of the line
    and end at the end of the line.

    :param sql_pattern: the string with the pattern in SQL syntax
    :return: a string with the pattern in regex syntax
    """

    def tokenizer(string, tokens_to_apply):
        """Recursive function that tokenizes a string using the provided tokens

        :param string: the string to tokenize
        :param tokens_to_apply: the list of tokens still to process (in order: the first will be processed first)
        :return: a tokenized and escaped string for regex
        """
        if tokens_to_apply:
            # We still have tokens to process
            # find the first occurrence of the first token passed in the list
            # note that the order of the tokens list is important, e.g. we need
            # to match first \% and then %
            first, sep, rest = string.partition(tokens_to_apply[0])

            # There is indeed a separator:
            if sep:
                # at least one token was found; therefore I have to map tokens[sep]
                # to the corresponding regex expression (via the dictionary substitution)
                # Moreover, the 'rest' is not empty so I apply recursively `tokenizer`,
                # with ALL tokens passed (there could be more occurrences of tokens_to_apply[0])
                # Instead, for the first part, we know that we found the FIRST occurrence of tokens_to_apply[0]
                # so I pass the list without the first element
                return (
                    tokenizer(first, tokens_to_apply=tokens_to_apply[1:])
                    + dict(SQL_TO_REGEX_TOKENS)[sep]
                    + tokenizer(rest, tokens_to_apply=tokens_to_apply)
                )
            # Here sep is empty: it means also rest is empty, and we just
            # return (recursively) the tokenizer on the first part, avoiding
            # infinite loops
            return tokenizer(first, tokens_to_apply=tokens_to_apply[1:])
        # There is no more token to process: we now have
        # a string that we know we want to consider literally, so
        # we just make sure it's escaped for a regex to avoid
        # that the user passed in weird characters that are valid
        # regex symbols like $ ^ [ ] etc
        return re.escape(string)

    return f'^{tokenizer(sql_pattern, tokens_to_apply=[token_pair[0] for token_pair in SQL_TO_REGEX_TOKENS])}$'


def sql_string_match(string, pattern):
    """Check if the string matches the provided pattern,
    specified using SQL syntax.

    See documentation of :py:func:`~aiida.common.escaping.get_regex_pattern_from_sql`
    for an explanation of the syntax.

    :param string: the string to check
    :param pattern: the SQL pattern
    :return: True if the string matches, False otherwise
    """
    return bool(re.match(get_regex_pattern_from_sql(pattern), string))



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Fix calculation states.

`UNDETERMINED` and `NOTFOUND` `dbcalcstate.state` values are replaced by `FAILED`.

Revision ID: django_0002
Revises: django_0001

"""

from alembic import op

revision = 'django_0002'
down_revision = 'django_0001'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    # Note in the original django migration, a warning log was actually added to the node,
    # but we forgo that here
    op.execute(
        """
        UPDATE db_dbcalcstate
        SET state = 'FAILED'
        WHERE state = 'NOTFOUND'
    """
    )
    op.execute(
        """
        UPDATE db_dbcalcstate
        SET state = 'FAILED'
        WHERE state = 'UNDETERMINED'
    """
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0002.')



"""Interface for links of a node instance."""

from __future__ import annotations

import typing as t
from typing import Optional, cast

from aiida.common import exceptions
from aiida.common.escaping import sql_string_match
from aiida.common.lang import type_check
from aiida.common.links import LinkType

from ..querybuilder import QueryBuilder
from ..utils.links import LinkManager, LinkTriple

if t.TYPE_CHECKING:
    from .node import Node


class NodeLinks:
    """Interface for links of a node instance."""

    def __init__(self, node: 'Node') -> None:
        """Initialize the links interface."""
        self._node = node
        self.incoming_cache: list[LinkTriple] = []

    def _add_incoming_cache(self, source: 'Node', link_type: LinkType, link_label: str) -> None:
        """Add an incoming link to the cache.

        .. note: the proposed link is not validated in this function, so this should not be called directly
            but it should only be called by `Node.add_incoming`.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :raise aiida.common.UniquenessError: if the given link triple already exists in the cache
        """
        assert self.incoming_cache is not None, 'incoming_cache not initialised'

        link_triple = LinkTriple(source, link_type, link_label)

        if link_triple in self.incoming_cache:
            raise exceptions.UniquenessError(f'the link triple {link_triple} is already present in the cache')

        self.incoming_cache.append(link_triple)

    def add_incoming(self, source: 'Node', link_type: LinkType, link_label: str) -> None:
        """Add a link of the given type from a given node to ourself.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :raise TypeError: if `source` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        """
        self.validate_incoming(source, link_type, link_label)
        source.base.links.validate_outgoing(self._node, link_type, link_label)

        if self._node.is_stored and source.is_stored:
            self._node.backend_entity.add_incoming(source.backend_entity, link_type, link_label)
        else:
            self._add_incoming_cache(source, link_type, link_label)

    def validate_incoming(self, source: 'Node', link_type: LinkType, link_label: str) -> None:
        """Validate adding a link of the given type from a given node to ourself.

        This function will first validate the types of the inputs, followed by the node and link types and validate
        whether in principle a link of that type between the nodes of these types is allowed.

        Subsequently, the validity of the "degree" of the proposed link is validated, which means validating the
        number of links of the given type from the given node type is allowed.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :raise TypeError: if `source` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        """
        from aiida.orm.utils.links import validate_link

        from .node import Node

        validate_link(source, self._node, link_type, link_label, backend=self._node.backend)

        # Check if the proposed link would introduce a cycle in the graph following ancestor/descendant rules
        if link_type in [LinkType.CREATE, LinkType.INPUT_CALC, LinkType.INPUT_WORK]:
            builder = (
                QueryBuilder(backend=self._node.backend)
                .append(Node, filters={'id': self._node.pk}, tag='parent')
                .append(Node, filters={'id': source.pk}, tag='child', with_ancestors='parent')
            )
            if builder.count() > 0:
                raise ValueError('the link you are attempting to create would generate a cycle in the graph')

    def validate_outgoing(self, target: 'Node', link_type: LinkType, link_label: str) -> None:
        """Validate adding a link of the given type from ourself to a given node.

        The validity of the triple (source, link, target) should be validated in the `validate_incoming` call.
        This method will be called afterwards and can be overriden by subclasses to add additional checks that are
        specific to that subclass.

        :param target: the node to which the link is going
        :param link_type: the link type
        :param link_label: the link label
        :raise TypeError: if `target` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        """
        from .node import Node

        type_check(link_type, LinkType, f'link_type should be a LinkType enum but got: {type(link_type)}')
        type_check(target, Node, f'target should be a `Node` instance but got: {type(target)}')

    def get_stored_link_triples(
        self,
        node_class: Optional[t.Type['Node']] = None,
        link_type: t.Union[LinkType, t.Sequence[LinkType]] = (),
        link_label_filter: t.Optional[str] = None,
        link_direction: str = 'incoming',
        only_uuid: bool = False,
    ) -> list[LinkTriple]:
        """Return the list of stored link triples directly incoming to or outgoing of this node.

        Note this will only return link triples that are stored in the database. Anything in the cache is ignored.

        :param node_class: If specified, should be a class, and it filters only elements of that (subclass of) type
        :param link_type: Only get inputs of this link type, if empty tuple then returns all inputs of all link types.
        :param link_label_filter: filters the incoming nodes by its link label. This should be a regex statement as
            one would pass directly to a QueryBuilder filter statement with the 'like' operation.
        :param link_direction: `incoming` or `outgoing` to get the incoming or outgoing links, respectively.
        :param only_uuid: project only the node UUID instead of the instance onto the `NodeTriple.node` entries
        """
        from .node import Node

        if not isinstance(link_type, (tuple, list)):
            link_type = cast(t.Sequence[LinkType], (link_type,))

        if link_type and not all(isinstance(t, LinkType) for t in link_type):
            raise TypeError(f'link_type should be a LinkType or tuple of LinkType: got {link_type}')

        node_class = node_class or Node
        node_filters: dict[str, t.Any] = {'id': {'==': self._node.pk}}
        edge_filters: dict[str, t.Any] = {}

        if link_type:
            edge_filters['type'] = {'in': [t.value for t in link_type]}

        if link_label_filter:
            edge_filters['label'] = {'like': link_label_filter}

        builder = QueryBuilder(backend=self._node.backend)
        builder.append(Node, filters=node_filters, tag='main')

        node_project = ['uuid'] if only_uuid else ['*']
        if link_direction == 'outgoing':
            builder.append(
                node_class,
                with_incoming='main',
                project=node_project,
                edge_project=['type', 'label'],
                edge_filters=edge_filters,
            )
        else:
            builder.append(
                node_class,
                with_outgoing='main',
                project=node_project,
                edge_project=['type', 'label'],
                edge_filters=edge_filters,
            )

        return [LinkTriple(entry[0], LinkType(entry[1]), entry[2]) for entry in builder.all()]

    def get_incoming(
        self,
        node_class: Optional[t.Type['Node']] = None,
        link_type: t.Union[LinkType, t.Sequence[LinkType]] = (),
        link_label_filter: t.Optional[str] = None,
        only_uuid: bool = False,
    ) -> LinkManager:
        """Return a list of link triples that are (directly) incoming into this node.

        :param node_class: If specified, should be a class or tuple of classes, and it filters only
            elements of that specific type (or a subclass of 'type')
        :param link_type: If specified should be a string or tuple to get the inputs of this
            link type, if None then returns all inputs of all link types.
        :param link_label_filter: filters the incoming nodes by its link label.
            Here wildcards (% and _) can be passed in link label filter as we are using "like" in QB.
        :param only_uuid: project only the node UUID instead of the instance onto the `NodeTriple.node` entries
        """
        if not isinstance(link_type, (tuple, list)):
            link_type = cast(t.Sequence[LinkType], (link_type,))

        if self._node.is_stored:
            link_triples = self.get_stored_link_triples(
                node_class, link_type, link_label_filter, 'incoming', only_uuid=only_uuid
            )
        else:
            link_triples = []

        # Get all cached link triples
        for link_triple in self.incoming_cache:
            if only_uuid:
                link_triple = LinkTriple(  # noqa: PLW2901
                    link_triple.node.uuid,  # type: ignore[arg-type]
                    link_triple.link_type,
                    link_triple.link_label,
                )

            if link_triple in link_triples:
                raise exceptions.InternalError(
                    f'Node<{self._node.pk}> has both a stored and cached link triple {link_triple}'
                )

            if not link_type or link_triple.link_type in link_type:
                if link_label_filter is not None:
                    if sql_string_match(string=link_triple.link_label, pattern=link_label_filter):
                        link_triples.append(link_triple)
                else:
                    link_triples.append(link_triple)

        return LinkManager(link_triples)

    def get_outgoing(
        self,
        node_class: Optional[t.Type['Node']] = None,
        link_type: t.Union[LinkType, t.Sequence[LinkType]] = (),
        link_label_filter: t.Optional[str] = None,
        only_uuid: bool = False,
    ) -> LinkManager:
        """Return a list of link triples that are (directly) outgoing of this node.

        :param node_class: If specified, should be a class or tuple of classes, and it filters only
            elements of that specific type (or a subclass of 'type')
        :param link_type: If specified should be a string or tuple to get the inputs of this
            link type, if None then returns all outputs of all link types.
        :param link_label_filter: filters the outgoing nodes by its link label.
            Here wildcards (% and _) can be passed in link label filter as we are using "like" in QB.
        :param only_uuid: project only the node UUID instead of the instance onto the `NodeTriple.node` entries
        """
        link_triples = self.get_stored_link_triples(node_class, link_type, link_label_filter, 'outgoing', only_uuid)
        return LinkManager(link_triples)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with `Node` sub class for workflow function processes."""

from typing import TYPE_CHECKING

from aiida.common.links import LinkType
from aiida.orm.utils.mixins import FunctionCalculationMixin

from .workflow import WorkflowNode, WorkflowNodeLinks

if TYPE_CHECKING:
    from aiida.orm import Node

__all__ = ('WorkFunctionNode',)


class WorkFunctionNodeLinks(WorkflowNodeLinks):
    """Interface for links of a node instance."""

    def validate_outgoing(self, target: 'Node', link_type: LinkType, link_label: str) -> None:
        """Validate adding a link of the given type from ourself to a given node.

        A workfunction cannot create Data, so if we receive an outgoing RETURN link to an unstored Data node, that means
        the user created a Data node within our function body and is trying to return it. This use case should be
        reserved for @calcfunctions, as they can have CREATE links.

        :param target: the node to which the link is going
        :param link_type: the link type
        :param link_label: the link label
        :raise TypeError: if `target` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        """
        super().validate_outgoing(target, link_type, link_label)
        if link_type is LinkType.RETURN and not target.is_stored:
            raise ValueError(
                'trying to return an unstored Data node from a @workfunction, however, @workfunctions cannot create '
                'data. You probably want to use a @calcfunction instead.'
            )


class WorkFunctionNode(FunctionCalculationMixin, WorkflowNode):  # type: ignore[misc]
    """ORM class for all nodes representing the execution of a workfunction."""

    _CLS_NODE_LINKS = WorkFunctionNodeLinks



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Move trajectory symbols from repository array to attribute

Note, this is similar to the django migration django_0026

Revision ID: 12536798d4d3
Revises: 37f3d4882837
Create Date: 2019-01-21 10:15:02.451308

"""

from alembic import op
from sqlalchemy import Integer, String, cast
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.sql import column, func, select, table

from aiida.storage.psql_dos.backend import get_filepath_container
from aiida.storage.psql_dos.migrations.utils.utils import load_numpy_array_from_repository

# revision identifiers, used by Alembic.
revision = '12536798d4d3'
down_revision = '37f3d4882837'
branch_labels = None
depends_on = None

# Here we duplicate the data stored in a TrajectoryData symbols array, storing it as an attribute.
# We delete the duplicates in the following migration (ce56d84bcc35) to avoid to delete data


def upgrade():
    """Migrations for the upgrade."""
    connection = op.get_bind()
    repo_path = get_filepath_container(op.get_context().opts['aiida_profile']).parent

    DbNode = table(  # noqa: N806
        'db_dbnode',
        column('id', Integer),
        column('uuid', UUID),
        column('type', String),
        column('attributes', JSONB),
    )

    nodes = connection.execute(
        select(DbNode.c.id, DbNode.c.uuid).where(
            DbNode.c.type == op.inline_literal('node.data.array.trajectory.TrajectoryData.')
        )
    ).fetchall()

    for pk, uuid in nodes:
        symbols = load_numpy_array_from_repository(repo_path, uuid, 'symbols').tolist()
        connection.execute(
            DbNode.update()
            .where(DbNode.c.id == pk)
            .values(
                attributes=func.jsonb_set(DbNode.c.attributes, op.inline_literal('{"symbols"}'), cast(symbols, JSONB))
            )
        )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of 12536798d4d3.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""" "Implementation of `DbImporter` for the NNIN/C database."""

from aiida.tools.dbimporters.baseclasses import DbImporter, DbSearchResults, UpfEntry


class NnincDbImporter(DbImporter):
    """Database importer for NNIN/C Pseudopotential Virtual Vault."""

    def _str_clause(self, key, alias, values):
        """Returns part of HTTP GET query for querying string fields."""
        if not isinstance(values, str):
            raise ValueError(f"incorrect value for keyword '{alias}' -- only strings and integers are accepted")
        return f'{key}={values}'

    _keywords = {
        'xc_approximation': ['frmxcprox', _str_clause],
        'xc_type': ['frmxctype', _str_clause],
        'pseudopotential_class': ['frmspclass', _str_clause],
        'element': ['element', None],
    }

    def __init__(self, **kwargs):
        self._query_url = 'http://nninc.cnf.cornell.edu/dd_search.php'
        self.setup_db(**kwargs)

    def query_get(self, **kwargs):
        """Forms a HTTP GET query for querying the NNIN/C Pseudopotential
        Virtual Vault.

        :return: a string with HTTP GET statement.
        """
        get_parts = []
        for key, value in self._keywords.items():
            if key in kwargs:
                values = kwargs.pop(key)
                if value[1] is not None:
                    get_parts.append(value[1](self, value[0], key, values))

        if kwargs:
            raise NotImplementedError(f"following keyword(s) are not implemented: {', '.join(kwargs.keys())}")

        return f"{self._query_url}?{'&'.join(get_parts)}"

    def query(self, **kwargs):
        """Performs a query on the NNIN/C Pseudopotential Virtual Vault using
        ``keyword = value`` pairs, specified in ``kwargs``.

        :return: an instance of
            :py:class:`aiida.tools.dbimporters.plugins.nninc.NnincSearchResults`.
        """
        import re
        from urllib.request import urlopen

        query = self.query_get(**kwargs)
        with urlopen(query) as handle:
            response = handle.read()
        results = re.findall(r'psp_files/([^\']+)\.UPF', response)

        elements = kwargs.get('element', None)
        if elements and not isinstance(elements, list):
            elements = [elements]

        if elements:
            results_now = set()
            for psp in results:
                for element in elements:
                    if psp.startswith(f'{element}.'):
                        results_now = results_now | set([psp])
            results = list(results_now)

        return NnincSearchResults([{'id': x} for x in results])

    def setup_db(self, query_url=None, **kwargs):
        """Changes the database connection details."""
        if query_url:
            self._query_url = query_url

        if kwargs.keys():
            raise NotImplementedError(
                "unknown database connection parameter(s): '"
                + "', '".join(kwargs.keys())
                + "', available parameters: 'query_url'"
            )

    def get_supported_keywords(self):
        """Returns the list of all supported query keywords.

        :return: list of strings
        """
        return self._keywords.keys()


class NnincSearchResults(DbSearchResults):
    """Results of the search, performed on NNIN/C Pseudopotential Virtual
    Vault.
    """

    _base_url = 'http://nninc.cnf.cornell.edu/psp_files/'

    def __init__(self, results):
        super().__init__(results)
        self._return_class = NnincEntry

    def __len__(self):
        return len(self._results)

    def _get_source_dict(self, result_dict):
        """Returns a dictionary, which is passed as kwargs to the created
        DbEntry instance, describing the source of the entry.

        :param result_dict: dictionary, describing an entry in the results.
        """
        return {'id': result_dict['id']}

    def _get_url(self, result_dict):
        """Returns an URL of an entry CIF file.

        :param result_dict: dictionary, describing an entry in the results.
        """
        return f"{self._base_url + result_dict['id']}.UPF"


class NnincEntry(UpfEntry):
    """Represents an entry from NNIN/C Pseudopotential Virtual Vault."""

    def __init__(self, uri, **kwargs):
        """Creates an instance of
        :py:class:`aiida.tools.dbimporters.plugins.nninc.NnincEntry`, related
        to the supplied URI.
        """
        super().__init__(
            db_name='NNIN/C Pseudopotential Virtual Vault', db_uri='http://nninc.cnf.cornell.edu', uri=uri, **kwargs
        )



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Base settings required for the configuration of an AiiDA instance."""

from __future__ import annotations

import os
import pathlib
import warnings

DEFAULT_UMASK = 0o0077
DEFAULT_AIIDA_PATH_VARIABLE = 'AIIDA_PATH'
DEFAULT_AIIDA_PATH = '~'
DEFAULT_AIIDA_USER = 'aiida@localhost'
DEFAULT_CONFIG_DIR_NAME = '.aiida'
DEFAULT_CONFIG_FILE_NAME = 'config.json'
DEFAULT_CONFIG_INDENT_SIZE = 4
DEFAULT_DAEMON_DIR_NAME = 'daemon'
DEFAULT_DAEMON_LOG_DIR_NAME = 'log'
DEFAULT_ACCESS_CONTROL_DIR_NAME = 'access'

# Assign defaults which may be overriden in set_configuration_directory() below
AIIDA_CONFIG_FOLDER: pathlib.Path = pathlib.Path(DEFAULT_AIIDA_PATH).expanduser() / DEFAULT_CONFIG_DIR_NAME
DAEMON_DIR: pathlib.Path = AIIDA_CONFIG_FOLDER / DEFAULT_DAEMON_DIR_NAME
DAEMON_LOG_DIR: pathlib.Path = DAEMON_DIR / DEFAULT_DAEMON_LOG_DIR_NAME
ACCESS_CONTROL_DIR: pathlib.Path = AIIDA_CONFIG_FOLDER / DEFAULT_ACCESS_CONTROL_DIR_NAME


def create_instance_directories() -> None:
    """Create the base directories required for a new AiiDA instance.

    This will create the base AiiDA directory defined by the AIIDA_CONFIG_FOLDER variable, unless it already exists.
    Subsequently, it will create the daemon directory within it and the daemon log directory.
    """
    from aiida.common import ConfigurationError

    directory_base = AIIDA_CONFIG_FOLDER.expanduser()
    directory_daemon = directory_base / DAEMON_DIR
    directory_daemon_log = directory_base / DAEMON_LOG_DIR
    directory_access = directory_base / ACCESS_CONTROL_DIR

    list_of_paths = [
        directory_base,
        directory_daemon,
        directory_daemon_log,
        directory_access,
    ]

    umask = os.umask(DEFAULT_UMASK)

    try:
        for path in list_of_paths:
            if path is directory_base and not path.exists():
                warnings.warn(f'Creating AiiDA configuration folder `{path}`.')

            try:
                path.mkdir(parents=True, exist_ok=True)
            except OSError as exc:
                raise ConfigurationError(f'could not create the `{path}` configuration directory: {exc}') from exc
    finally:
        os.umask(umask)


def get_configuration_directory():
    """Return the path of the configuration directory.

    The location of the configuration directory is defined following these heuristics in order:

        * If the ``AIIDA_PATH`` variable is set, all the paths will be checked to see if they contain a
          configuration folder. The first one to be encountered will be set as ``AIIDA_CONFIG_FOLDER``. If none of them
          contain one, the last path defined in the environment variable considered is used.
        * If an existing directory is still not found, the ``DEFAULT_AIIDA_PATH`` is used.

    :returns: The path of the configuration directory.
    """
    dirpath_config = get_configuration_directory_from_envvar()

    # If no existing configuration directory is found, fall back to the default
    if dirpath_config is None:
        dirpath_config = pathlib.Path(DEFAULT_AIIDA_PATH).expanduser() / DEFAULT_CONFIG_DIR_NAME

    return dirpath_config


def get_configuration_directory_from_envvar() -> pathlib.Path | None:
    """Return the path of a config directory from the ``AIIDA_PATH`` environment variable.

    The environment variable should be a colon separated string of filepaths that either point directly to a config
    directory or a path that contains a config directory. The first match is returned. If no existing config directory
    is found, the last path in the environment variable is used.

    :returns: The path of the configuration directory or ``None`` if ``AIIDA_PATH`` is not defined.
    """
    environment_variable = os.environ.get(DEFAULT_AIIDA_PATH_VARIABLE)

    if environment_variable is None:
        return None

    # Loop over all the paths in the ``AIIDA_PATH`` variable to see if any of them contain a configuration folder
    for base_dir_path in [path for path in environment_variable.split(':') if path]:
        dirpath_config = pathlib.Path(base_dir_path).expanduser()

        # Only add the base config directory name to the base path if it does not already do so
        # Someone might already include it in the environment variable. e.g.: ``AIIDA_PATH=/home/some/path/.aiida``
        if dirpath_config.name != DEFAULT_CONFIG_DIR_NAME:
            dirpath_config = dirpath_config / DEFAULT_CONFIG_DIR_NAME

        # If the directory exists, we leave it set and break the loop
        if dirpath_config.is_dir():
            break

    return dirpath_config


def set_configuration_directory(aiida_config_folder: pathlib.Path | None = None) -> None:
    """Set the configuration directory, related global variables and create instance directories.

    The location of the configuration directory is defined by ``aiida_config_folder`` or if not defined, the path that
    is returned by ``get_configuration_directory``. If the directory does not exist yet, it is created, together with
    all its subdirectories.
    """
    global AIIDA_CONFIG_FOLDER  # noqa: PLW0603
    global DAEMON_DIR  # noqa: PLW0603
    global DAEMON_LOG_DIR  # noqa: PLW0603
    global ACCESS_CONTROL_DIR  # noqa: PLW0603

    AIIDA_CONFIG_FOLDER = aiida_config_folder or get_configuration_directory()
    DAEMON_DIR = AIIDA_CONFIG_FOLDER / DEFAULT_DAEMON_DIR_NAME
    DAEMON_LOG_DIR = DAEMON_DIR / DEFAULT_DAEMON_LOG_DIR_NAME
    ACCESS_CONTROL_DIR = AIIDA_CONFIG_FOLDER / DEFAULT_ACCESS_CONTROL_DIR_NAME

    create_instance_directories()


# Initialize the configuration directory settings
set_configuration_directory()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Legacy migrations,
using the old ``data.json`` format for storing the database.

These migrations simply manipulate the metadata and data in-place.
"""

from typing import Callable, Dict, Tuple

from .v04_to_v05 import migrate_v4_to_v5
from .v05_to_v06 import migrate_v5_to_v6
from .v06_to_v07 import migrate_v6_to_v7
from .v07_to_v08 import migrate_v7_to_v8
from .v08_to_v09 import migrate_v8_to_v9
from .v09_to_v10 import migrate_v9_to_v10
from .v10_to_v11 import migrate_v10_to_v11
from .v11_to_v12 import migrate_v11_to_v12
from .v12_to_v13 import migrate_v12_to_v13

# version from -> version to, function which modifies metadata, data in-place
LEGACY_MIGRATE_FUNCTIONS: Dict[str, Tuple[str, Callable[[dict, dict], None]]] = {
    '0.4': ('0.5', migrate_v4_to_v5),
    '0.5': ('0.6', migrate_v5_to_v6),
    '0.6': ('0.7', migrate_v6_to_v7),
    '0.7': ('0.8', migrate_v7_to_v8),
    '0.8': ('0.9', migrate_v8_to_v9),
    '0.9': ('0.10', migrate_v9_to_v10),
    '0.10': ('0.11', migrate_v10_to_v11),
    '0.11': ('0.12', migrate_v11_to_v12),
    '0.12': ('0.13', migrate_v12_to_v13),
}
FINAL_LEGACY_VERSION = '0.13'



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Change type string for `Data` nodes, from `data.*` to `node.data.*`

Note, this is identical to sqlalchemy migration 6a5c2ea1439d

Revision ID: django_0025
Revises: django_0024

"""

import sqlalchemy as sa
from alembic import op

revision = 'django_0025'
down_revision = 'django_0024'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    # The type string for `Data` nodes changed from `data.*` to `node.data.*`.
    statement = sa.text(
        r"""
        UPDATE db_dbnode
        SET type = regexp_replace(type, '^data.', 'node.data.')
        WHERE type LIKE 'data.%'
    """
    )
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0025.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Import an archive."""

from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, Literal, Optional, Set, Tuple, Union

from tabulate import tabulate

from aiida import orm
from aiida.common import timezone
from aiida.common.exceptions import IncompatibleStorageSchema
from aiida.common.lang import type_check
from aiida.common.links import LinkType
from aiida.common.log import AIIDA_LOGGER
from aiida.common.progress_reporter import get_progress_reporter
from aiida.manage import get_manager
from aiida.orm.entities import EntityTypes
from aiida.orm.implementation import StorageBackend
from aiida.orm.querybuilder import QueryBuilder
from aiida.repository import Repository

from .abstract import ArchiveFormatAbstract
from .common import batch_iter, entity_type_to_orm
from .exceptions import ImportTestRun, ImportUniquenessError, ImportValidationError
from .implementations.sqlite_zip import ArchiveFormatSqlZip

__all__ = ('IMPORT_LOGGER', 'import_archive')

IMPORT_LOGGER = AIIDA_LOGGER.getChild('export')

MergeExtrasType = Tuple[Literal['k', 'n'], Literal['c', 'n'], Literal['l', 'u', 'd']]
MergeExtraDescs = (
    {'k': '(k)eep', 'n': 'do (n)ot keep'},
    {'c': '(c)reate', 'n': 'do (n)ot create'},
    {'l': '(l)eave existing', 'u': '(u)pdate with new', 'd': '(d)elete'},
)
MergeCommentsType = Literal['leave', 'newest', 'overwrite']

DUPLICATE_LABEL_MAX = 100
DUPLICATE_LABEL_TEMPLATE = '{0} (Imported #{1})'


@dataclass
class QueryParams:
    """Parameters for executing backend queries."""

    batch_size: int
    """Batch size for streaming database rows."""
    filter_size: int
    """Maximum number of parameters allowed in a single query filter."""


def import_archive(
    path: Union[str, Path],
    *,
    archive_format: Optional[ArchiveFormatAbstract] = None,
    filter_size: int = 999,
    batch_size: int = 1000,
    import_new_extras: bool = True,
    merge_extras: MergeExtrasType = ('k', 'n', 'l'),
    merge_comments: MergeCommentsType = 'leave',
    include_authinfos: bool = False,
    create_group: bool = True,
    group: Optional[orm.Group] = None,
    test_run: bool = False,
    backend: Optional[StorageBackend] = None,
) -> Optional[int]:
    """Import an archive into the AiiDA backend.

    :param path: the path to the archive
    :param archive_format: The class for interacting with the archive
    :param filter_size: Maximum size of parameters allowed in a single query filter
    :param batch_size: Batch size for streaming database rows
    :param import_new_extras: Keep extras on new nodes (except private aiida keys), else strip
    :param merge_extras: Rules for merging extras into existing nodes.
        The first letter acts on extras that are present in the original node and not present in the imported node.
        Can be either:
        'k' (keep it) or
        'n' (do not keep it).
        The second letter acts on the imported extras that are not present in the original node.
        Can be either:
        'c' (create it) or
        'n' (do not create it).
        The third letter defines what to do in case of a name collision.
        Can be either:
        'l' (leave the old value),
        'u' (update with a new value),
        'd' (delete the extra)
    :param create_group: Add all imported nodes to the specified group, or an automatically created one
    :param group: Group wherein all imported Nodes will be placed.
        If None, one will be auto-generated.
    :param test_run: if True, do not write to file
    :param backend: the backend to import to. If not specified, the default backend is used.

    :returns: Primary Key of the import Group

    :raises `~aiida.common.exceptions.CorruptStorage`: if the provided archive cannot be read.
    :raises `~aiida.common.exceptions.IncompatibleStorageSchema`: if the archive version is not at head.
    :raises `~aiida.tools.archive.exceptions.ImportValidationError`: if invalid entities are found in the archive.
    :raises `~aiida.tools.archive.exceptions.ImportUniquenessError`: if a new unique entity can not be created.
    """
    archive_format = archive_format or ArchiveFormatSqlZip()
    type_check(path, (str, Path))
    type_check(archive_format, ArchiveFormatAbstract)
    type_check(batch_size, int)
    type_check(import_new_extras, bool)
    type_check(merge_extras, tuple)
    if len(merge_extras) != 3:
        raise ValueError('merge_extras not of length 3')
    if not (merge_extras[0] in ['k', 'n'] and merge_extras[1] in ['c', 'n'] and merge_extras[2] in ['l', 'u', 'd']):
        raise ValueError('merge_extras contains invalid values')
    if merge_comments not in ('leave', 'newest', 'overwrite'):
        raise ValueError(f"merge_comments not in {('leave', 'newest', 'overwrite')!r}")
    type_check(group, orm.Group, allow_none=True)
    type_check(test_run, bool)
    backend = backend or get_manager().get_profile_storage()
    type_check(backend, StorageBackend)
    query_params = QueryParams(batch_size=batch_size, filter_size=filter_size)

    if group and not group.is_stored:
        group.store()

    # check the version is latest
    # to-do we should have a way to check the version against aiida-core
    # i.e. its not whether the version is the latest that matters, it is that it is compatible with the backend version
    # its a bit weird at the moment because django/sqlalchemy have different versioning
    if not archive_format.read_version(path) == archive_format.latest_version:
        raise IncompatibleStorageSchema(
            f'The archive version {archive_format.read_version(path)!r} '
            f'is not the latest version {archive_format.latest_version!r}'
        )

    IMPORT_LOGGER.report(
        str(
            tabulate(
                [
                    ['Archive', Path(path).name],
                    ['New Node Extras', 'keep' if import_new_extras else 'strip'],
                    ['Merge Node Extras (in database)', MergeExtraDescs[0][merge_extras[0]]],
                    ['Merge Node Extras (in archive)', MergeExtraDescs[1][merge_extras[1]]],
                    ['Merge Node Extras (in both)', MergeExtraDescs[2][merge_extras[2]]],
                    ['Merge Comments', merge_comments],
                    ['Computer Authinfos', 'include' if include_authinfos else 'exclude'],
                ],
                headers=['Parameters', ''],
            )
        )
        + '\n'
    )

    if test_run:
        IMPORT_LOGGER.report('Test run: nothing will be added to the profile')

    with archive_format.open(path, mode='r') as reader:
        backend_from = reader.get_backend()

        # To ensure we do not corrupt the backend database on a faulty import,
        # Every addition/update is made in a single transaction, which is commited on exit
        with backend.transaction():
            user_ids_archive_backend = _import_users(backend_from, backend, query_params)
            computer_ids_archive_backend = _import_computers(backend_from, backend, query_params)
            if include_authinfos:
                _import_authinfos(
                    backend_from, backend, query_params, user_ids_archive_backend, computer_ids_archive_backend
                )
            node_ids_archive_backend = _import_nodes(
                backend_from,
                backend,
                query_params,
                user_ids_archive_backend,
                computer_ids_archive_backend,
                import_new_extras,
                merge_extras,
            )
            _import_logs(backend_from, backend, query_params, node_ids_archive_backend)
            _import_comments(
                backend_from, backend, query_params, user_ids_archive_backend, node_ids_archive_backend, merge_comments
            )
            _import_links(backend_from, backend, query_params, node_ids_archive_backend)
            group_labels = _import_groups(
                backend_from, backend, query_params, user_ids_archive_backend, node_ids_archive_backend
            )
            import_group_id = None
            if create_group:
                import_group_id = _make_import_group(
                    group, group_labels, node_ids_archive_backend, backend, query_params
                )
            new_repo_keys = _get_new_object_keys(archive_format.key_format, backend_from, backend, query_params)

            if test_run:
                # exit before we write anything to the database or repository
                raise ImportTestRun('test run complete')

            # now the transaction has been successfully populated, but not committed, we add the repository files
            # if the commit fails, this is not so much an issue, since the files can be removed on repo maintenance
            _add_files_to_repo(backend_from, backend, new_repo_keys)

            IMPORT_LOGGER.report('Committing transaction to database...')

    return import_group_id


def _add_new_entities(
    etype: EntityTypes,
    total: int,
    unique_field: str,
    backend_unique_id: dict,
    backend_from: StorageBackend,
    backend_to: StorageBackend,
    query_params: QueryParams,
    transform: Callable[[dict], dict],
) -> None:
    """Add new entities to the output backend and update the mapping of unique field -> id."""
    IMPORT_LOGGER.report(f'Adding {total} new {etype.value}(s)')

    # collect the unique entities from the input backend to be added to the output backend
    ufields = []
    query = QueryBuilder(backend=backend_from).append(entity_type_to_orm[etype], project=unique_field)
    for (ufield,) in query.distinct().iterall(batch_size=query_params.batch_size):
        if ufield not in backend_unique_id:
            ufields.append(ufield)

    with get_progress_reporter()(desc=f'Adding new {etype.value}(s)', total=total) as progress:
        # batch the filtering of rows by filter size, to limit the number of query variables used in any one query,
        # since certain backends have a limit on the number of variables in a query (such as SQLITE_MAX_VARIABLE_NUMBER)
        for nrows, ufields_batch in batch_iter(ufields, query_params.filter_size):
            rows = [
                transform(row)
                for row in QueryBuilder(backend=backend_from)
                .append(
                    entity_type_to_orm[etype],
                    filters={unique_field: {'in': ufields_batch}},
                    project=['**'],
                    tag='entity',
                )
                .dict(batch_size=query_params.batch_size)
            ]
            new_ids = backend_to.bulk_insert(etype, rows)
            backend_unique_id.update({row[unique_field]: pk for pk, row in zip(new_ids, rows)})
            progress.update(nrows)


def _import_users(
    backend_from: StorageBackend, backend_to: StorageBackend, query_params: QueryParams
) -> Dict[int, int]:
    """Import users from one backend to another.

    :returns: mapping of input backend id to output backend id
    """
    # get the records from the input backend
    qbuilder = QueryBuilder(backend=backend_from)
    input_id_email = dict(qbuilder.append(orm.User, project=['id', 'email']).all(batch_size=query_params.batch_size))

    # get matching emails from the backend
    output_email_id: Dict[str, int] = {}
    if input_id_email:
        output_email_id = dict(
            orm.QueryBuilder(backend=backend_to)
            .append(orm.User, filters={'email': {'in': list(input_id_email.values())}}, project=['email', 'id'])
            .all(batch_size=query_params.batch_size)
        )

    new_users = len(input_id_email) - len(output_email_id)
    existing_users = len(output_email_id)

    if existing_users:
        IMPORT_LOGGER.report(f'Skipping {existing_users} existing User(s)')
    if new_users:
        # add new users and update output_email_id with their email -> id mapping
        def transform(row):
            return {k: v for k, v in row['entity'].items() if k != 'id'}

        _add_new_entities(
            EntityTypes.USER, new_users, 'email', output_email_id, backend_from, backend_to, query_params, transform
        )

    # generate mapping of input backend id to output backend id
    return {int(i): output_email_id[email] for i, email in input_id_email.items()}


def _import_computers(
    backend_from: StorageBackend, backend_to: StorageBackend, query_params: QueryParams
) -> Dict[int, int]:
    """Import computers from one backend to another.

    :returns: mapping of input backend id to output backend id
    """
    # get the records from the input backend
    qbuilder = QueryBuilder(backend=backend_from)
    input_id_uuid = dict(qbuilder.append(orm.Computer, project=['id', 'uuid']).all(batch_size=query_params.batch_size))

    # get matching uuids from the backend
    backend_uuid_id: Dict[str, int] = {}
    if input_id_uuid:
        backend_uuid_id = dict(
            orm.QueryBuilder(backend=backend_to)
            .append(orm.Computer, filters={'uuid': {'in': list(input_id_uuid.values())}}, project=['uuid', 'id'])
            .all(batch_size=query_params.batch_size)
        )

    new_computers = len(input_id_uuid) - len(backend_uuid_id)
    existing_computers = len(backend_uuid_id)

    if existing_computers:
        IMPORT_LOGGER.report(f'Skipping {existing_computers} existing Computer(s)')
    if new_computers:
        # add new computers and update backend_uuid_id with their uuid -> id mapping

        # Labels should be unique, so we create new labels on clashes
        labels = {
            label
            for (label,) in orm.QueryBuilder(backend=backend_to)
            .append(orm.Computer, project='label')
            .iterall(batch_size=query_params.batch_size)
        }
        relabelled = 0

        def transform(row: dict) -> dict:
            data = row['entity']
            pk = data.pop('id')
            nonlocal labels
            if data['label'] in labels:
                for i in range(DUPLICATE_LABEL_MAX):
                    new_label = DUPLICATE_LABEL_TEMPLATE.format(data['label'], i)
                    if new_label not in labels:
                        data['label'] = new_label
                        break
                else:
                    raise ImportUniquenessError(
                        f'Archive Computer {pk} has existing label {data["label"]!r} and re-labelling failed'
                    )
                nonlocal relabelled
                relabelled += 1
            labels.add(data['label'])
            return data

        _add_new_entities(
            EntityTypes.COMPUTER,
            new_computers,
            'uuid',
            backend_uuid_id,
            backend_from,
            backend_to,
            query_params,
            transform,
        )

        if relabelled:
            IMPORT_LOGGER.report(f'Re-labelled {relabelled} new Computer(s)')

    # generate mapping of input backend id to output backend id
    return {int(i): backend_uuid_id[uuid] for i, uuid in input_id_uuid.items()}


def _import_authinfos(
    backend_from: StorageBackend,
    backend_to: StorageBackend,
    query_params: QueryParams,
    user_ids_archive_backend: Dict[int, int],
    computer_ids_archive_backend: Dict[int, int],
) -> None:
    """Import logs from one backend to another.

    :returns: mapping of input backend id to output backend id
    """
    # get the records from the input backend
    qbuilder = QueryBuilder(backend=backend_from)
    input_id_user_comp = qbuilder.append(
        orm.AuthInfo,
        project=['id', 'aiidauser_id', 'dbcomputer_id'],
    ).all(batch_size=query_params.batch_size)

    # translate user_id / computer_id, from -> to
    try:
        to_user_id_comp_id = [
            (user_ids_archive_backend[_user_id], computer_ids_archive_backend[_comp_id])
            for _, _user_id, _comp_id in input_id_user_comp
        ]
    except KeyError as exception:
        raise ImportValidationError(f'Archive AuthInfo has unknown User/Computer: {exception}')

    # retrieve existing user_id / computer_id
    backend_id_user_comp = []
    if to_user_id_comp_id:
        qbuilder = orm.QueryBuilder(backend=backend_to)
        qbuilder.append(
            orm.AuthInfo,
            filters={
                'aiidauser_id': {'in': [_user_id for _user_id, _ in to_user_id_comp_id]},
                'dbcomputer_id': {'in': [_comp_id for _, _comp_id in to_user_id_comp_id]},
            },
            project=['id', 'aiidauser_id', 'dbcomputer_id'],
        )
        backend_id_user_comp = [
            (user_id, comp_id)
            for _, user_id, comp_id in qbuilder.all(batch_size=query_params.batch_size)
            if (user_id, comp_id) in to_user_id_comp_id
        ]

    new_authinfos = len(input_id_user_comp) - len(backend_id_user_comp)
    existing_authinfos = len(backend_id_user_comp)

    if existing_authinfos:
        IMPORT_LOGGER.report(f'Skipping {existing_authinfos} existing AuthInfo(s)')
    if not new_authinfos:
        return

    # import new authinfos
    IMPORT_LOGGER.report(f'Adding {new_authinfos} new {EntityTypes.AUTHINFO.value}(s)')
    new_ids = [
        _id
        for _id, _user_id, _comp_id in input_id_user_comp
        if (user_ids_archive_backend[_user_id], computer_ids_archive_backend[_comp_id]) not in backend_id_user_comp
    ]
    qbuilder = QueryBuilder(backend=backend_from).append(
        orm.AuthInfo, filters={'id': {'in': new_ids}}, project=['**'], tag='entity'
    )
    iterator = qbuilder.iterdict()

    def transform(row: dict) -> dict:
        data = row['entity']
        data.pop('id')
        data['aiidauser_id'] = user_ids_archive_backend[data['aiidauser_id']]
        data['dbcomputer_id'] = computer_ids_archive_backend[data['dbcomputer_id']]
        return data

    with get_progress_reporter()(
        desc=f'Adding new {EntityTypes.AUTHINFO.value}(s)', total=qbuilder.count()
    ) as progress:
        for nrows, rows in batch_iter(iterator, query_params.batch_size, transform):
            backend_to.bulk_insert(EntityTypes.AUTHINFO, rows)
            progress.update(nrows)


def _import_nodes(
    backend_from: StorageBackend,
    backend_to: StorageBackend,
    query_params: QueryParams,
    user_ids_archive_backend: Dict[int, int],
    computer_ids_archive_backend: Dict[int, int],
    import_new_extras: bool,
    merge_extras: MergeExtrasType,
) -> Dict[int, int]:
    """Import nodes from one backend to another.

    :returns: mapping of input backend id to output backend id
    """
    IMPORT_LOGGER.report('Collecting Node(s) ...')
    # get the records from the input backend
    qbuilder = QueryBuilder(backend=backend_from)
    input_id_uuid = dict(qbuilder.append(orm.Node, project=['id', 'uuid']).all(batch_size=query_params.batch_size))

    # get matching uuids from the backend
    backend_uuid_id: Dict[str, int] = {}
    input_id_uuid_uuids = list(input_id_uuid.values())

    if input_id_uuid:
        for _, batch in batch_iter(input_id_uuid_uuids, query_params.filter_size):
            backend_uuid_id.update(
                dict(
                    orm.QueryBuilder(backend=backend_to)
                    .append(orm.Node, filters={'uuid': {'in': batch}}, project=['uuid', 'id'])
                    .all(batch_size=query_params.batch_size)
                )
            )

    new_nodes = len(input_id_uuid) - len(backend_uuid_id)

    if backend_uuid_id:
        _merge_node_extras(backend_from, backend_to, query_params, backend_uuid_id, merge_extras)

    if new_nodes:
        # add new nodes and update backend_uuid_id with their uuid -> id mapping
        transform = NodeTransform(user_ids_archive_backend, computer_ids_archive_backend, import_new_extras)
        _add_new_entities(
            EntityTypes.NODE, new_nodes, 'uuid', backend_uuid_id, backend_from, backend_to, query_params, transform
        )

    # generate mapping of input backend id to output backend id
    return {int(i): backend_uuid_id[uuid] for i, uuid in input_id_uuid.items()}


class NodeTransform:
    """Callable to transform a Node DB row, between the source archive and target backend."""

    def __init__(
        self,
        user_ids_archive_backend: Dict[int, int],
        computer_ids_archive_backend: Dict[int, int],
        import_new_extras: bool,
    ):
        """Construct a new instance."""
        self.user_ids_archive_backend = user_ids_archive_backend
        self.computer_ids_archive_backend = computer_ids_archive_backend
        self.import_new_extras = import_new_extras

    def __call__(self, row: dict) -> dict:
        """Perform the transform."""
        data = row['entity']
        pk = data.pop('id')
        try:
            data['user_id'] = self.user_ids_archive_backend[data['user_id']]
        except KeyError as exc:
            raise ImportValidationError(f'Archive Node {pk} has unknown User: {exc}')
        if data['dbcomputer_id'] is not None:
            try:
                data['dbcomputer_id'] = self.computer_ids_archive_backend[data['dbcomputer_id']]
            except KeyError as exc:
                raise ImportValidationError(f'Archive Node {pk} has unknown Computer: {exc}')
        if self.import_new_extras:
            # Remove node hashing and other aiida "private" extras
            data['extras'] = {k: v for k, v in data['extras'].items() if not k.startswith('_aiida_')}
            if data.get('node_type', '').endswith('code.Code.'):
                data['extras'].pop('hidden', None)
        else:
            data['extras'] = {}
        if data.get('node_type', '').startswith('process.'):
            # remove checkpoint from attributes of process nodes
            data['attributes'].pop(orm.ProcessNode.CHECKPOINT_KEY, None)
        return data


def _import_logs(
    backend_from: StorageBackend,
    backend_to: StorageBackend,
    query_params: QueryParams,
    node_ids_archive_backend: Dict[int, int],
) -> Dict[int, int]:
    """Import logs from one backend to another.

    :returns: mapping of input backend id to output backend id
    """
    # get the records from the input backend
    qbuilder = QueryBuilder(backend=backend_from)
    input_id_uuid = dict(qbuilder.append(orm.Log, project=['id', 'uuid']).all(batch_size=query_params.batch_size))

    # get matching uuids from the backend
    backend_uuid_id: Dict[str, int] = {}
    if input_id_uuid:
        backend_uuid_id = dict(
            orm.QueryBuilder(backend=backend_to)
            .append(orm.Log, filters={'uuid': {'in': list(input_id_uuid.values())}}, project=['uuid', 'id'])
            .all(batch_size=query_params.batch_size)
        )

    new_logs = len(input_id_uuid) - len(backend_uuid_id)
    existing_logs = len(backend_uuid_id)

    if existing_logs:
        IMPORT_LOGGER.report(f'Skipping {existing_logs} existing Log(s)')
    if new_logs:
        # add new logs and update backend_uuid_id with their uuid -> id mapping
        def transform(row: dict) -> dict:
            data = row['entity']
            pk = data.pop('id')
            try:
                data['dbnode_id'] = node_ids_archive_backend[data['dbnode_id']]
            except KeyError as exc:
                raise ImportValidationError(f'Archive Log {pk} has unknown Node: {exc}')
            return data

        _add_new_entities(
            EntityTypes.LOG, new_logs, 'uuid', backend_uuid_id, backend_from, backend_to, query_params, transform
        )

    # generate mapping of input backend id to output backend id
    return {int(i): backend_uuid_id[uuid] for i, uuid in input_id_uuid.items()}


def _merge_node_extras(
    backend_from: StorageBackend,
    backend_to: StorageBackend,
    query_params: QueryParams,
    backend_uuid_id: Dict[str, int],
    mode: MergeExtrasType,
) -> None:
    """Merge extras from the input backend with the ones in the output backend.

    :param backend_uuid_id: mapping of uuid to output backend id
    :param mode: tuple of merge modes for extras
    """
    num_existing = len(backend_uuid_id)

    if mode == ('k', 'n', 'l'):
        # 'none': keep old extras, do not add imported ones
        IMPORT_LOGGER.report(f'Skipping {num_existing} existing Node(s)')
        return

    input_extras = (
        QueryBuilder(backend=backend_from)
        .append(
            orm.Node, tag='node', filters={'uuid': {'in': list(backend_uuid_id.keys())}}, project=['uuid', 'extras']
        )
        .order_by([{'node': 'uuid'}])
    )

    if mode == ('n', 'c', 'u'):
        # 'mirror' operation: remove old extras, put only the new ones
        IMPORT_LOGGER.report(f'Replacing {num_existing} existing Node extras')

        def transform(row):
            return {'id': backend_uuid_id[row[0]], 'extras': row[1]}

        with get_progress_reporter()(desc='Replacing extras', total=input_extras.count()) as progress:
            for nrows, rows in batch_iter(
                input_extras.iterall(batch_size=query_params.batch_size), query_params.batch_size, transform
            ):
                backend_to.bulk_update(EntityTypes.NODE, rows)
                progress.update(nrows)
        return

    # run (slower) generic merge operation
    backend_extras = (
        QueryBuilder(backend=backend_to)
        .append(
            orm.Node, tag='node', filters={'uuid': {'in': list(backend_uuid_id.keys())}}, project=['uuid', 'extras']
        )
        .order_by([{'node': 'uuid'}])
    )

    IMPORT_LOGGER.report(f'Merging {num_existing} existing Node extras')

    if not input_extras.count() == backend_extras.count():
        raise ImportValidationError(
            f'Number of Nodes in archive ({input_extras.count()}) and backend ({backend_extras.count()}) do not match'
        )

    def _transform(data: Tuple[Tuple[str, dict], Tuple[str, dict]]) -> dict:
        """Transform the new and existing extras into a dict that can be passed to bulk_update."""
        new_uuid, new_extras = data[0]
        old_uuid, old_extras = data[1]
        if new_uuid != old_uuid:
            raise ImportValidationError(f'UUID mismatch when merging node extras: {new_uuid} != {old_uuid}')
        backend_id = backend_uuid_id[new_uuid]
        old_keys = set(old_extras.keys())
        new_keys = set(new_extras.keys())
        collided_keys = old_keys.intersection(new_keys)
        old_keys_only = old_keys.difference(collided_keys)
        new_keys_only = new_keys.difference(collided_keys)

        final_extras = {}

        if mode == ('k', 'c', 'u'):
            # 'update_existing' operation: if an extra already exists,
            # overwrite its new value with a new one
            final_extras = new_extras
            for key in old_keys_only:
                final_extras[key] = old_extras[key]
            return {'id': backend_id, 'extras': final_extras}

        if mode == ('k', 'c', 'l'):
            # 'keep_existing': if an extra already exists, keep its original value
            final_extras = old_extras
            for key in new_keys_only:
                final_extras[key] = new_extras[key]
            return {'id': backend_id, 'extras': final_extras}

        if mode[0] == 'k':
            for key in old_keys_only:
                final_extras[key] = old_extras[key]
        elif mode[0] != 'n':
            raise ImportValidationError(
                f"Unknown first letter of the update extras mode: '{mode}'. Should be either 'k' or 'n'"
            )
        if mode[1] == 'c':
            for key in new_keys_only:
                final_extras[key] = new_extras[key]
        elif mode[1] != 'n':
            raise ImportValidationError(
                f"Unknown second letter of the update extras mode: '{mode}'. Should be either 'c' or 'n'"
            )
        if mode[2] == 'u':
            for key in collided_keys:
                final_extras[key] = new_extras[key]
        elif mode[2] == 'l':
            for key in collided_keys:
                final_extras[key] = old_extras[key]
        elif mode[2] != 'd':
            raise ImportValidationError(
                f"Unknown third letter of the update extras mode: '{mode}'. Should be one of 'u'/'l'/'a'/'d'"
            )
        return {'id': backend_id, 'extras': final_extras}

    with get_progress_reporter()(desc='Merging extras', total=input_extras.count()) as progress:
        for nrows, rows in batch_iter(
            zip(
                input_extras.iterall(batch_size=query_params.batch_size),
                backend_extras.iterall(batch_size=query_params.batch_size),
            ),
            query_params.batch_size,
            _transform,
        ):
            backend_to.bulk_update(EntityTypes.NODE, rows)
            progress.update(nrows)


class CommentTransform:
    """Callable to transform a Comment DB row, between the source archive and target backend."""

    def __init__(
        self,
        user_ids_archive_backend: Dict[int, int],
        node_ids_archive_backend: Dict[int, int],
    ):
        """Construct a new instance."""
        self.user_ids_archive_backend = user_ids_archive_backend
        self.node_ids_archive_backend = node_ids_archive_backend

    def __call__(self, row: dict) -> dict:
        """Perform the transform."""
        data = row['entity']
        pk = data.pop('id')
        try:
            data['user_id'] = self.user_ids_archive_backend[data['user_id']]
        except KeyError as exc:
            raise ImportValidationError(f'Archive Comment {pk} has unknown User: {exc}')
        try:
            data['dbnode_id'] = self.node_ids_archive_backend[data['dbnode_id']]
        except KeyError as exc:
            raise ImportValidationError(f'Archive Comment {pk} has unknown Node: {exc}')
        return data


def _import_comments(
    backend_from: StorageBackend,
    backend: StorageBackend,
    query_params: QueryParams,
    user_ids_archive_backend: Dict[int, int],
    node_ids_archive_backend: Dict[int, int],
    merge_comments: MergeCommentsType,
) -> Dict[int, int]:
    """Import comments from one backend to another.

    :returns: mapping of archive id to backend id
    """
    # get the records from the input backend
    qbuilder = QueryBuilder(backend=backend_from)
    input_id_uuid = dict(qbuilder.append(orm.Comment, project=['id', 'uuid']).all(batch_size=query_params.batch_size))

    # get matching uuids from the backend
    backend_uuid_id: Dict[str, int] = {}
    if input_id_uuid:
        backend_uuid_id = dict(
            orm.QueryBuilder(backend=backend)
            .append(orm.Comment, filters={'uuid': {'in': list(input_id_uuid.values())}}, project=['uuid', 'id'])
            .all(batch_size=query_params.batch_size)
        )

    new_comments = len(input_id_uuid) - len(backend_uuid_id)
    existing_comments = len(backend_uuid_id)

    archive_comments = QueryBuilder(backend=backend_from).append(
        orm.Comment, filters={'uuid': {'in': list(backend_uuid_id.keys())}}, project=['uuid', 'mtime', 'content']
    )

    if existing_comments:
        if merge_comments == 'leave':
            IMPORT_LOGGER.report(f'Skipping {existing_comments} existing Comment(s)')
        elif merge_comments == 'overwrite':
            IMPORT_LOGGER.report(f'Overwriting {existing_comments} existing Comment(s)')

            def _transform(row):
                data = {'id': backend_uuid_id[row[0]], 'mtime': row[1], 'content': row[2]}
                return data

            with get_progress_reporter()(desc='Overwriting comments', total=archive_comments.count()) as progress:
                for nrows, rows in batch_iter(
                    archive_comments.iterall(batch_size=query_params.batch_size), query_params.batch_size, _transform
                ):
                    backend.bulk_update(EntityTypes.COMMENT, rows)
                    progress.update(nrows)

        elif merge_comments == 'newest':
            IMPORT_LOGGER.report(f'Updating {existing_comments} existing Comment(s)')

            def _transform(row):
                # to-do this is probably not the most efficient way to do this
                uuid, new_mtime, new_comment = row
                cmt = orm.comments.CommentCollection(orm.Comment, backend).get(uuid=uuid)
                if cmt.mtime < new_mtime:
                    cmt.set_mtime(new_mtime)
                    cmt.set_content(new_comment)

            with get_progress_reporter()(desc='Updating comments', total=archive_comments.count()) as progress:
                for nrows, rows in batch_iter(
                    archive_comments.iterall(batch_size=query_params.batch_size), query_params.batch_size, _transform
                ):
                    progress.update(nrows)

        else:
            raise ImportValidationError(f'Unknown merge_comments value: {merge_comments}.')
    if new_comments:
        # add new comments and update backend_uuid_id with their uuid -> id mapping
        _add_new_entities(
            EntityTypes.COMMENT,
            new_comments,
            'uuid',
            backend_uuid_id,
            backend_from,
            backend,
            query_params,
            CommentTransform(user_ids_archive_backend, node_ids_archive_backend),
        )

    # generate mapping of input backend id to output backend id
    return {int(i): backend_uuid_id[uuid] for i, uuid in input_id_uuid.items()}


def _import_links(
    backend_from: StorageBackend,
    backend_to: StorageBackend,
    query_params: QueryParams,
    node_ids_archive_backend: Dict[int, int],
) -> None:
    """Import links from one backend to another."""
    # initial variables
    calculation_node_types = 'process.calculation.'
    workflow_node_types = 'process.workflow.'
    data_node_types = 'data.'
    allowed_link_nodes = {
        LinkType.CALL_CALC: (workflow_node_types, calculation_node_types),
        LinkType.CALL_WORK: (workflow_node_types, workflow_node_types),
        LinkType.CREATE: (calculation_node_types, data_node_types),
        LinkType.INPUT_CALC: (data_node_types, calculation_node_types),
        LinkType.INPUT_WORK: (data_node_types, workflow_node_types),
        LinkType.RETURN: (workflow_node_types, data_node_types),
    }
    link_type_uniqueness = {
        LinkType.CALL_CALC: ('out_id',),
        LinkType.CALL_WORK: ('out_id',),
        LinkType.CREATE: (
            'in_id_label',
            'out_id',
        ),
        LinkType.INPUT_CALC: ('out_id_label',),
        LinkType.INPUT_WORK: ('out_id_label',),
        LinkType.RETURN: ('in_id_label',),
    }

    # Batch by type, to reduce memory load
    # to-do check no extra types in archive?
    for link_type in LinkType:
        # get validation parameters
        allowed_in_type, allowed_out_type = allowed_link_nodes[link_type]
        link_uniqueness = link_type_uniqueness[link_type]

        # count links of this type in archive
        archive_query = (
            QueryBuilder(backend=backend_from)
            .append(orm.Node, tag='incoming', project=['id', 'node_type'])
            .append(
                orm.Node,
                with_incoming='incoming',
                project=['id', 'node_type'],
                edge_filters={'type': link_type.value},
                edge_project=['id', 'label'],
            )
        )
        total = archive_query.count()

        if not total:
            continue  # nothing to add

        # get existing links set, to check existing
        IMPORT_LOGGER.report(f'Gathering existing {link_type.value!r} Link(s)')
        existing_links = {
            tuple(link)
            for link in orm.QueryBuilder(backend=backend_to)
            .append(entity_type='link', filters={'type': link_type.value}, project=['input_id', 'output_id', 'label'])
            .iterall(batch_size=query_params.batch_size)
        }
        # create additional validators
        # note, we only populate them when required, to reduce memory usage
        existing_in_id_label = (
            {(link[0], link[2]) for link in existing_links} if 'in_id_label' in link_uniqueness else set()
        )
        existing_out_id = {link[1] for link in existing_links} if 'out_id' in link_uniqueness else set()
        existing_out_id_label = (
            {(link[1], link[2]) for link in existing_links} if 'out_id_label' in link_uniqueness else set()
        )

        # loop through archive links; validate and add new
        new_count = existing_count = 0
        insert_rows = []
        with get_progress_reporter()(desc=f'Processing {link_type.value!r} Link(s)', total=total) as progress:
            for in_id, in_type, out_id, out_type, link_id, link_label in archive_query.iterall(
                batch_size=query_params.batch_size
            ):
                progress.update()

                # convert ids: archive -> profile
                try:
                    in_id = node_ids_archive_backend[in_id]  # noqa: PLW2901
                except KeyError as exc:
                    raise ImportValidationError(f'Archive Link {link_id} has unknown input Node: {exc}')
                try:
                    out_id = node_ids_archive_backend[out_id]  # noqa: PLW2901
                except KeyError as exc:
                    raise ImportValidationError(f'Archive Link {link_id} has unknown output Node: {exc}')

                # skip existing links
                if (in_id, out_id, link_label) in existing_links:
                    existing_count += 1
                    continue

                # validation
                if in_id == out_id:
                    raise ImportValidationError(f'Cannot add a link to oneself: {in_id}')
                if not in_type.startswith(allowed_in_type):
                    raise ImportValidationError(
                        f'Cannot add a {link_type.value!r} link from {in_type} (link {link_id})'
                    )
                if not out_type.startswith(allowed_out_type):
                    raise ImportValidationError(f'Cannot add a {link_type.value!r} link to {out_type} (link {link_id})')
                if 'in_id_label' in link_uniqueness and (in_id, link_label) in existing_in_id_label:
                    raise ImportUniquenessError(
                        f'Node {in_id} already has an outgoing {link_type.value!r} link with label {link_label!r}'
                    )
                if 'out_id' in link_uniqueness and out_id in existing_out_id_label:
                    raise ImportUniquenessError(f'Node {out_id} already has an incoming {link_type.value!r} link')
                if 'out_id_label' in link_uniqueness and (out_id, link_label) in existing_out_id_label:
                    raise ImportUniquenessError(
                        f'Node {out_id} already has an incoming {link_type.value!r} link with label {link_label!r}'
                    )

                # update variables
                new_count += 1
                insert_rows.append(
                    {
                        'input_id': in_id,
                        'output_id': out_id,
                        'type': link_type.value,
                        'label': link_label,
                    }
                )
                existing_links.add((in_id, out_id, link_label))
                existing_in_id_label.add((in_id, link_label))
                existing_out_id.add(out_id)
                existing_out_id_label.add((out_id, link_label))

                # flush new rows, once batch size is reached
                if (new_count % query_params.batch_size) == 0:
                    backend_to.bulk_insert(EntityTypes.LINK, insert_rows)
                    insert_rows = []

            # flush remaining new rows
            if insert_rows:
                backend_to.bulk_insert(EntityTypes.LINK, insert_rows)

        # report counts
        if existing_count:
            IMPORT_LOGGER.report(f'Skipped {existing_count} existing {link_type.value!r} Link(s)')
        if new_count:
            IMPORT_LOGGER.report(f'Added {new_count} new {link_type.value!r} Link(s)')


class GroupTransform:
    """Callable to transform a Group DB row, between the source archive and target backend."""

    def __init__(self, user_ids_archive_backend: Dict[int, int], labels: Set[str]):
        self.user_ids_archive_backend = user_ids_archive_backend
        self.labels = labels
        self.relabelled = 0

    def __call__(self, row: dict) -> dict:
        """Perform the transform."""
        data = row['entity']
        pk = data.pop('id')
        try:
            data['user_id'] = self.user_ids_archive_backend[data['user_id']]
        except KeyError as exc:
            raise ImportValidationError(f'Archive Group {pk} has unknown User: {exc}')
        # Labels should be unique, so we create new labels on clashes
        if data['label'] in self.labels:
            for i in range(DUPLICATE_LABEL_MAX):
                new_label = DUPLICATE_LABEL_TEMPLATE.format(data['label'], i)
                if new_label not in self.labels:
                    data['label'] = new_label
                    break
            else:
                raise ImportUniquenessError(
                    f'Archive Group {pk} has existing label {data["label"]!r} and re-labelling failed'
                )
            self.relabelled += 1
        self.labels.add(data['label'])
        return data


def _import_groups(
    backend_from: StorageBackend,
    backend_to: StorageBackend,
    query_params: QueryParams,
    user_ids_archive_backend: Dict[int, int],
    node_ids_archive_backend: Dict[int, int],
) -> Set[str]:
    """Import groups from the input backend, and add group -> node records.

    :returns: Set of labels
    """
    # get the records from the input backend
    qbuilder = QueryBuilder(backend=backend_from)
    input_id_uuid = dict(qbuilder.append(orm.Group, project=['id', 'uuid']).all(batch_size=query_params.batch_size))

    # get matching uuids from the backend
    backend_uuid_id: Dict[str, int] = {}
    if input_id_uuid:
        backend_uuid_id = dict(
            orm.QueryBuilder(backend=backend_to)
            .append(orm.Group, filters={'uuid': {'in': list(input_id_uuid.values())}}, project=['uuid', 'id'])
            .all(batch_size=query_params.batch_size)
        )

    # get all labels
    labels = {
        label
        for (label,) in orm.QueryBuilder(backend=backend_to)
        .append(orm.Group, project='label')
        .iterall(batch_size=query_params.batch_size)
    }

    new_groups = len(input_id_uuid) - len(backend_uuid_id)
    new_uuids = set(input_id_uuid.values()).difference(backend_uuid_id.keys())
    existing_groups = len(backend_uuid_id)

    if existing_groups:
        IMPORT_LOGGER.report(f'Skipping {existing_groups} existing Group(s)')
    if new_groups:
        # add new groups and update backend_uuid_id with their uuid -> id mapping

        transform = GroupTransform(user_ids_archive_backend, labels)

        _add_new_entities(
            EntityTypes.GROUP, new_groups, 'uuid', backend_uuid_id, backend_from, backend_to, query_params, transform
        )

        if transform.relabelled:
            IMPORT_LOGGER.report(f'Re-labelled {transform.relabelled} new Group(s)')

        # generate mapping of input backend id to output backend id
        group_id_archive_backend = {i: backend_uuid_id[uuid] for i, uuid in input_id_uuid.items()}
        # Add nodes to new groups
        iterator = (
            QueryBuilder(backend=backend_from)
            .append(orm.Group, project='id', filters={'uuid': {'in': new_uuids}}, tag='group')
            .append(orm.Node, project='id', with_group='group')
        )
        total = iterator.count()
        if total:
            IMPORT_LOGGER.report(f'Adding {total} Node(s) to new Group(s)')

            def group_node_transform(row):
                group_id = group_id_archive_backend[row[0]]
                try:
                    node_id = node_ids_archive_backend[row[1]]
                except KeyError as exc:
                    raise ImportValidationError(f'Archive Group {group_id} has unknown Node: {exc}')
                return {'dbgroup_id': group_id, 'dbnode_id': node_id}

            with get_progress_reporter()(desc=f'Adding new {EntityTypes.GROUP_NODE.value}(s)', total=total) as progress:
                for nrows, rows in batch_iter(
                    iterator.iterall(batch_size=query_params.batch_size), query_params.batch_size, group_node_transform
                ):
                    backend_to.bulk_insert(EntityTypes.GROUP_NODE, rows)
                    progress.update(nrows)

    return labels


def _make_import_group(
    group: Optional[orm.Group],
    labels: Set[str],
    node_ids_archive_backend: Dict[int, int],
    backend_to: StorageBackend,
    query_params: QueryParams,
) -> Optional[int]:
    """Make an import group containing all imported nodes.

    :param group: Use an existing group
    :param labels: All existing group labels on the backend
    :param node_ids_archive_backend: node pks to add to the group

    :returns: The id of the group

    """
    # Do not create an empty group
    if not node_ids_archive_backend:
        IMPORT_LOGGER.debug('No nodes to import, so no import group created')
        return None

    # Get the Group id
    if group is None:
        # Get an unique name for the import group, based on the current (local) time
        label = timezone.localtime(timezone.now()).strftime('%Y%m%d-%H%M%S')
        if label in labels:
            for i in range(DUPLICATE_LABEL_MAX):
                new_label = DUPLICATE_LABEL_TEMPLATE.format(label, i)
                if new_label not in labels:
                    label = new_label
                    break
            else:
                raise ImportUniquenessError(f'New import Group has existing label {label!r} and re-labelling failed')
        dummy_orm = orm.ImportGroup(label, backend=backend_to)
        row = {
            'label': label,
            'description': 'Group generated by archive import',
            'type_string': dummy_orm.type_string,
            'user_id': dummy_orm.user.pk,
        }
        (group_id,) = backend_to.bulk_insert(EntityTypes.GROUP, [row], allow_defaults=True)
        IMPORT_LOGGER.report(f'Created new import Group: PK={group_id}, label={label}')
        group_node_ids = set()
    else:
        group_id = group.pk  # type: ignore[assignment]
        IMPORT_LOGGER.report(f'Using existing import Group: PK={group_id}, label={group.label}')
        group_node_ids = {
            pk
            for (pk,) in orm.QueryBuilder(backend=backend_to)
            .append(orm.Group, filters={'id': group_id}, tag='group')
            .append(orm.Node, with_group='group', project='id')
            .iterall(batch_size=query_params.batch_size)
        }

    # Add all the nodes to the Group
    with get_progress_reporter()(
        desc='Adding all Node(s) to the import Group', total=len(node_ids_archive_backend)
    ) as progress:
        iterator = (
            {'dbgroup_id': group_id, 'dbnode_id': node_id}
            for node_id in node_ids_archive_backend.values()
            if node_id not in group_node_ids
        )
        for nrows, rows in batch_iter(iterator, query_params.batch_size):
            backend_to.bulk_insert(EntityTypes.GROUP_NODE, rows)
            progress.update(nrows)

    return group_id


def _get_new_object_keys(
    key_format: str, backend_from: StorageBackend, backend_to: StorageBackend, query_params: QueryParams
) -> Set[str]:
    """Return the object keys that need to be added to the backend."""
    archive_hashkeys: Set[str] = set()
    query = QueryBuilder(backend=backend_from).append(orm.Node, project='repository_metadata')
    with get_progress_reporter()(desc='Collecting archive Node file keys', total=query.count()) as progress:
        for (repository_metadata,) in query.iterall(batch_size=query_params.batch_size):
            archive_hashkeys.update(key for key in Repository.flatten(repository_metadata).values() if key is not None)
            progress.update()

    IMPORT_LOGGER.report('Checking keys against repository ...')

    repository = backend_to.get_repository()
    if not repository.key_format == key_format:
        raise NotImplementedError(
            f'Backend repository key format incompatible: {repository.key_format!r} != {key_format!r}'
        )
    new_hashkeys = archive_hashkeys.difference(repository.list_objects())

    existing_count = len(archive_hashkeys) - len(new_hashkeys)
    if existing_count:
        IMPORT_LOGGER.report(f'Skipping {existing_count} existing repository files')
    if new_hashkeys:
        IMPORT_LOGGER.report(f'Adding {len(new_hashkeys)} new repository files')

    return new_hashkeys


def _add_files_to_repo(backend_from: StorageBackend, backend_to: StorageBackend, new_keys: Set[str]) -> None:
    """Add the new files to the repository."""
    if not new_keys:
        return None

    repository_to = backend_to.get_repository()
    repository_from = backend_from.get_repository()
    with get_progress_reporter()(desc='Adding archive files to repository', total=len(new_keys)) as progress:
        for key, handle in repository_from.iter_object_streams(new_keys):  # type: ignore[arg-type]
            backend_key = repository_to.put_object_from_filelike(handle)
            if backend_key != key:
                raise ImportValidationError(
                    f'Archive repository key is different to backend key: {key!r} != {backend_key!r}'
                )
            progress.update()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""A monkey-patched subclass of click.Option that does not evaluate callable default during tab completion."""

import typing as t

import click

__all__ = ('CallableDefaultOption',)


class CallableDefaultOption(click.Option):
    """A monkeypatch for click.Option that does not evaluate default callbacks during tab completion

    This is a temporary solution until a proper fix is implemented in click, see:
    https://github.com/pallets/click/issues/2614
    """

    def get_default(self, ctx: click.Context, call: bool = True) -> t.Optional[t.Union[t.Any, t.Callable[[], t.Any]]]:
        """Return default unless in tab-completion context."""
        if ctx.resilient_parsing:
            return None
        return super().get_default(ctx=ctx, call=call)
